/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 * GapsCovers.C - implementacao dos algoritmos para calculo do menor circulo
 *                envolvente e do maior circulo vazio de um conjunto de pontos.
 * 
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    14.05.92  09:30 	Welson R. Jacometti  	Arquivo criado
 *
 */

#include "GapsCovers.h"
#include "Segment2D.h"
#include "Circle.h"
#include "Polygon.h"
#include "Diameter.h"
#include "voronoi.h"
#include "ConvexHull.h"

/*
 *********** variaveis para animacao ********/

static GraphicSegment2D *anim_s;
static GraphicCircle *anim_c;
static GraphicQuadEdge *anim_v;

#define anim_segment	((Segment2D *)anim_s->obj)
#define anim_circle	((Circle *)anim_c->obj)
#define anim_voronoi	((QuadEdge *)anim_v->obj)

void gaps_covers_animation_init(World *w) {
    static bool FIRST = true;
    
    if (FIRST) {
      anim_s = new GraphicSegment2D();
      anim_c = new GraphicCircle();
      anim_v = new GraphicQuadEdge();
      FIRST = false;
    }
    anim_s->set_all(w, FOREGROUND, BACKGROUND, 3, 0, false);
    anim_c->set_all(w, FOREGROUND, BACKGROUND, 0, 0, false);
    anim_v->set_all(w, FOREGROUND, BACKGROUND, 0, 0, false);
}

typedef QuadEdge VorDiagram;
typedef QuadEdge_ptr VorDiagram_ptr; 
typedef QuadEdge FVorDiagram;
typedef QuadEdge_ptr FVorDiagram_ptr;

#ifndef _Circle_ptr_list
#define _Circle_ptr_list
declare(list,Circle_ptr)
#endif ~_Circle_ptr_list;

#ifndef _Segment2D_ptr_list
#define _Segment2D_ptr_list
declare(list,Segment2D_ptr)
#endif ~_Segment2D_ptr_list;

list(Circle_ptr) *rec_circle_list;

QuadEdge_forall_bf_BEGIN(recursive_get_all_circles,e)
    if (!already && !(e->org()->at_infinity()) && !(e->org()->obj_volatile())) {
      rec_circle_list->push(new Circle(*e->org(), e->org()->e_distance(*(e->rot()->org()))));
      already = true;
    }
QuadEdge_forall_bf_END(recursive_get_all_circles,e)

list(Circle_ptr)* get_all_circles( FVorDiagram_ptr &fvor) {
    rec_circle_list = new list(Circle_ptr);
    if (fvor) {
      QEdge_ptr e = fvor->root();
      rec_flag = *(e->oflag());
      recursive_get_all_circles(e);
    }

    return rec_circle_list;
}

/*
 * Smallest Enclosing Circle (Shamos)
 *
 *   Implementacao do algoritmo proposto por Shamos e que Toussaint provou
 *   ser correto, apesar da afirmacao falsa de que o diametro de um conjunto
 *   de pontos e' uma das arestas (a maior) do dual do diagrama de Voronoi
 *   de vizinho mais distante. O resultado e' interessante pois, segundo
 *   Toussaint, se o menor circulo envolvente for dado por dois pontos (i.e.,
 *   o diametro) entao CERTAMENTE o diametro tera' uma aresta no dual do fvor e,
 *   nos casos onde o diametro nao determina uma aresta, CERTAMENTE o
 *   menor circulo envolvente e' dado por tres pontos.
 *
 */
 
Circle_ptr smallest_enclosing_circle_shamos(list(Point2D_ptr) &points, World *w, int animation_depth, int reference_depth) {
    Point2D_ptr p;
    Circle res;
    list(Circle_ptr) *fvor_circles;
    list(Segment2D_ptr) *dual_fvor_edges;
    FVorDiagram_ptr fvor;
    FVorDiagram dual_fvor;
    bool is_diameter = true;

    if (points.size() > 1) { 
      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        gaps_covers_animation_init(w);
      ANIMATION_END(w, animation_depth, reference_depth)
  
      fvor = farthest_neighbor_voronoi_diagram(points, w, animation_depth+1, reference_depth);
  
      dual_fvor.set(fvor->root()->rot(), 0, 0);	// extract information from
      dual_fvor_edges = dual_fvor.edges();	// the already constructed
      dual_fvor.set(0, 0, 0);			// dual (triangulation)
  
      Segment2D_ptr s;
      Segment2D gs = *(dual_fvor_edges->head());
      double max_length = gs.P1().e_distance_sqr(gs.P2());
    
      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        anim_segment->set(gs.P1(), gs.P2());
        anim_s->draw(w, XorMode);
        frame_stop_mark(w);
      ANIMATION_END(w, animation_depth, reference_depth)
    
      forall(s, *dual_fvor_edges) {
        double test = s->P1().e_distance_sqr(s->P2());
        if (test > max_length) {
  
          ANIMATION_BEGIN(w, animation_depth, reference_depth)
            anim_segment->set(gs.P1(), gs.P2());
            anim_s->draw(w, XorMode);
            anim_segment->set(s->P1(), s->P2());
            anim_s->draw(w, XorMode);
            frame_stop_mark(w);
          ANIMATION_END(w, animation_depth, reference_depth)
    
          max_length = test;
          gs = *s;
        }
        delete s;
      }

      delete dual_fvor_edges;

      // gs is the greatest edge in the dual of the fvor(points)

      res.set(middle(gs.P1(), gs.P2()), gs.P1().e_distance(gs.P2()) / 2);
    
      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        anim_circle->set(res.center(), res.radius());
        anim_c->draw(w, XorMode);
        frame_stop_mark(w);
      ANIMATION_END(w, animation_depth, reference_deth)
    
      forall(p, points) {

        ANIMATION_BEGIN(w, animation_depth, reference_depth)
          anim_circle->set(p->X(), p->Y(), 10);
          anim_c->draw(w, XorMode);
          frame_stop_mark(w);
          anim_c->draw(w, XorMode);
        ANIMATION_END(w, animation_depth, reference_depth)
      
        if (!res.inside(*p)) {
          is_diameter = false;
          points.init_iterator();
          break;
        }
      }

      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        anim_circle->set(res.center(), res.radius());
        anim_c->draw(w, XorMode);
      ANIMATION_END(w, animation_depth, reference_deth)

      if (!is_diameter) {
        Circle_ptr c;
        fvor_circles = get_all_circles(fvor);
        delete fvor;
        res = *(fvor_circles->head());

        forall(c, *fvor_circles) {      
    
          ANIMATION_BEGIN(w, animation_depth, reference_depth)
            anim_circle->set(c->center(), c->radius());
            anim_c->draw(w, XorMode);
            frame_stop_mark(w);
            anim_c->draw(w, XorMode);
          ANIMATION_END(w, animation_depth, reference_deth)
    
          if (c->radius() < res.radius())
            res = *c;
          delete c;
        }
  
        delete fvor_circles;
      }

      return (Circle_ptr)res.reproduct();
    }
    else
      return 0;
}    
      	
/*
 * Smallest Enclosing Circle
 *
 *   Implementation of the algorithm proposed by Toussaint
 *
 */
 
Circle_ptr smallest_enclosing_circle_toussaint(list(Point2D_ptr) &points, World *w, int animation_depth, int reference_depth) {
    Point2D_ptr p;
    Segment2D_ptr diameter;
    Circle res;
    list(Circle_ptr) *fvor_circles;
    list(Point2D_ptr) points_aux;
    Polygon_ptr hull;
    bool is_diameter = true;

    if (points.size() > 1) { 
      points_aux = points;
    
      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        gaps_covers_animation_init(w);
      ANIMATION_END(w, animation_depth, reference_depth)
       
      hull = new Polygon(graham_scan(points_aux, w, animation_depth+1, reference_depth));
      diameter = diameter_convex_polygon(hull, w, animation_depth+1, reference_depth);
      res.set(middle(diameter->P1(), diameter->P2()), diameter->P1().e_distance(diameter->P2()) / 2);
      delete diameter;
    
      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        anim_circle->set(res.center(), res.radius());
        anim_c->draw(w, XorMode);
        frame_stop_mark(w);
      ANIMATION_END(w, animation_depth, reference_deth)
    
      forall(p, points) {

        ANIMATION_BEGIN(w, animation_depth, reference_depth)
          anim_circle->set(p->X(), p->Y(), 10);
          anim_c->draw(w, XorMode);
          frame_stop_mark(w);
          anim_c->draw(w, XorMode);
        ANIMATION_END(w, animation_depth, reference_depth)
      
        if (!res.inside(*p)) {
          is_diameter = false;
          points.init_iterator();
          break;
        }
      }
    
      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        anim_circle->set(res.center(), res.radius());
        anim_c->draw(w, XorMode);
      ANIMATION_END(w, animation_depth, reference_deth)
    
      if (!is_diameter) {
        Circle_ptr c;
        FVorDiagram_ptr fvor = farthest_neighbor_voronoi_diagram(hull->vertices(), w, animation_depth+1, reference_depth);
        fvor_circles = get_all_circles(fvor);
        delete fvor;
        res = *(fvor_circles->head());
 
        forall(c, *fvor_circles) {      
    
          ANIMATION_BEGIN(w, animation_depth, reference_depth)
            anim_circle->set(c->center(), c->radius());
            anim_c->draw(w, XorMode);
            frame_stop_mark(w);
            anim_c->draw(w, XorMode);
          ANIMATION_END(w, animation_depth, reference_deth)
    
          if (c->radius() < res.radius())
            res = *c;
          delete c;
        }

        delete fvor_circles;
      }

      delete hull;    
      return (Circle_ptr)res.reproduct();
    }
    else
      return 0;
}

Circle_ptr circle_at( Point2D &c,  Point2D &limit) {
    return new Circle(c, c.e_distance(limit));
}

Circle_ptr largest_empty_circle(list(Point2D_ptr) &points, World *w, int animation_depth, int reference_depth) {
    VorDiagram_ptr vor; 
    list(Segment2D_ptr) hull;
    list(Circle_ptr) *by_vertices, by_intersections;
    list_item seg, seg_next;
    bool seg_intersection, seg_next_intersection;
    Point2D p_aux, p_seg, p_seg_next;
    QEdge_ptr e, eaux;
    Segment2D_ptr s;
    Circle_ptr c;
    Circle res;
    
    if (points.size() > 1) {
    	
      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        gaps_covers_animation_init(w);
      ANIMATION_END(w, animation_depth, reference_depth)

      vor = voronoi_diagram(points, w, animation_depth+1, reference_depth);

      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        anim_voronoi->set(vor->root(), vor->le(), vor->re());
        anim_v->draw(w, XorMode);
        frame_stop_mark(w);
      ANIMATION_END(w, animation_depth, reference_depth)
            
      // extract the convex hull from the Voronoi Diagram
    
      eaux = e = vor->root()->rot()->sym();
      do {
        hull.append(new Segment2D(*eaux->org(), *eaux->dest()));
        eaux = eaux->rprev();
      } while (eaux != e && eaux != e->sym());

      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        forall(s, hull) {
          anim_segment->set(s->P1(), s->P2());
          anim_s->draw(w, XorMode);
        }
        frame_stop_mark(w);
      ANIMATION_END(w, animation_depth, reference_depth)

      // find first intersection with the convex hull

      seg = hull.first();
      eaux = e = vor->root();
      do {
      	
        ANIMATION_BEGIN(w, animation_depth, reference_depth);
          anim_segment->set(*eaux->org(), *eaux->dest());
          anim_s->draw(w, XorMode);
          frame_stop_mark(w);
          anim_s->draw(w, XorMode);
        ANIMATION_END(w, animation_depth, reference_depth);

        if (Segment2D(*eaux->org(), *eaux->dest()).intersect(*hull[seg]))
          break;

        eaux = eaux->rprev();
      } while (true);

      // calculate first intersection and the corresponding circle
        
      p_aux = Segment2D(*eaux->org(), *eaux->dest()).intersection(*hull[seg]);
      by_intersections.append(circle_at(p_aux, *eaux->rot()->org()));
      
      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        anim_circle->set(p_aux, 6);
        anim_c->draw(w, XorMode);
        frame_stop_mark(w);  
      ANIMATION_END(w, animation_depth, reference_depth)

      // find all intersections with the convex hull
      
      e = eaux;
      eaux = eaux->sym()->lnext();
      seg_next = hull.cyclic_succ(seg);
      do {
        seg_intersection = seg_next_intersection = false;
          
        ANIMATION_BEGIN(w, animation_depth, reference_depth);
          anim_segment->set(*eaux->org(), *eaux->dest());
          anim_s->draw(w, XorMode);
          frame_stop_mark(w);
          anim_s->draw(w, XorMode);
        ANIMATION_END(w, animation_depth, reference_depth);

        if (Segment2D(*eaux->org(), *eaux->dest()).intersect(*hull[seg])) {
          p_seg = Segment2D(*eaux->org(), *eaux->dest()).intersection(*hull[seg]);
          seg_intersection = true;
        }        

        if (Segment2D(*eaux->org(), *eaux->dest()).intersect(*hull[seg_next])) {
          p_seg_next = Segment2D(*eaux->org(), *eaux->dest()).intersection(*hull[seg_next]);
          seg_next_intersection = true;
        } 

        eaux->org()->obj_volatile(true);
           
        // A instrucao acima demarca quais os vertices que podem ser considerados, 
        // ja' que vertices externos `a envoltoria nao contam. O uso deste recurso
        // intrinseco ao GeoLab nao tem efeitos colaterais pois a estrutura sera' 
        // desalocada antes do final do algoritmo.
                  
        if (seg_intersection || seg_next_intersection) {
          bool advance_seg = false;
          
          if (seg_intersection && seg_next_intersection) {
            double d1, d2;
            if (eaux->org()->at_infinity()) {
              d1 = eaux->dest()->e_distance_sqr(p_seg);
              d2 = eaux->dest()->e_distance_sqr(p_seg_next);
              if (d1 > d2)
                p_aux = p_seg;
              else {
              	p_aux = p_seg_next;
              	advance_seg = true;
              }
            }
            else {
              d1 = eaux->org()->e_distance_sqr(p_seg);
              d2 = eaux->org()->e_distance_sqr(p_seg_next);
              if (d1 < d2)
                p_aux = p_seg;
              else {
              	p_aux = p_seg_next;
              	advance_seg = true;
              }
            }
          }
          else
            if (seg_intersection)
              p_aux = p_seg;
            else {
              p_aux = p_seg_next;
              advance_seg = true;
            }      
            
          by_intersections.append(circle_at(p_aux, *eaux->rot()->org()));

          ANIMATION_BEGIN(w, animation_depth, reference_depth)
            anim_circle->set(p_aux, 6);
            anim_c->draw(w, XorMode);
            frame_stop_mark(w);  
          ANIMATION_END(w, animation_depth, reference_depth)

          if (advance_seg) {          
            seg = seg_next;
            seg_next = hull.cyclic_succ(seg_next);
          }
          
          eaux = eaux->sym();
        }

        eaux = eaux->lnext();
      } while (eaux != e);

      // get circles from vertices of the voronoi diagram internal to the hull

      by_vertices = get_all_circles(vor);
      
      // select largest circle (from by_intersections and by_vertices)
            
      res = *(by_intersections.head());
      forall(c, by_intersections) {      
    
        ANIMATION_BEGIN(w, animation_depth, reference_depth)
          anim_circle->set(c->center(), c->radius());
          anim_c->draw(w, XorMode);
          frame_stop_mark(w);
          anim_c->draw(w, XorMode);
        ANIMATION_END(w, animation_depth, reference_deth)
    
        if (c->radius() > res.radius())
          res = *c;
        delete c;
      }

      forall(c, *by_vertices) {      
    
        ANIMATION_BEGIN(w, animation_depth, reference_depth)
          anim_circle->set(c->center(), c->radius());
          anim_c->draw(w, XorMode);
          frame_stop_mark(w);
          anim_c->draw(w, XorMode);
        ANIMATION_END(w, animation_depth, reference_deth)
    
        if (c->radius() > res.radius())
          res = *c;
        delete c;
      }

      delete vor;
      forall(s, hull) {
        delete s;
      }      
      hull.clear();
      by_intersections.clear();
      delete by_vertices;
      
      return (Circle_ptr)res.reproduct();
    }
    else    
      return 0;
}

// Modelagem

static list(Point2D_ptr) extract_points(const SelectedObjects &SO,const  GraphicObjects &GO) {
    list(Point2D_ptr) res;
    list_item it;
    forall(it, SO) {			 	// is_able_to_handle deve ser
      res.push((Point2D *)GO[it]->obj);		// chamado antes para certificar
    }						// que sao todos pontos
    return res;
}

bool GapsSetOfPoints::is_able_to_handle(const SelectedObjects &SO, const GraphicObjects &GO) {
    list_item it;
    forall(it, SO) {
      if (GO[it]->type() != Point2D_t) {
      	SO.init_iterator();
      	return false;
      }
    }
    
    return true;
}

ResultObjects SmallestEnclosingCircleShamos::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    Circle_ptr c;
        
    if (!SO.empty()) {
      c = smallest_enclosing_circle_shamos(extract_points(SO, GO), w, animation_depth, reference_depth);
      if (c) result.push((Obj_ptr)c);
    }
      
    return result;
}
    
ResultObjects SmallestEnclosingCircleToussaint::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    Circle_ptr c;
        
    if (!SO.empty()) {
      c = smallest_enclosing_circle_toussaint(extract_points(SO, GO), w, animation_depth, reference_depth);
      if (c) result.push((Obj_ptr)c);
    }
      
    return result;
}
    
ResultObjects LargestEmptyCircle::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    Circle_ptr c;
        
    if (!SO.empty()) {
      c = largest_empty_circle(extract_points(SO, GO), w, animation_depth, reference_depth);
      if (c) result.push((Obj_ptr)c);
    }
      
    return result;
}
    
Algorithm* smallest_enclosing_circle_shamos_link() {
    return new SmallestEnclosingCircleShamos();
}

Algorithm* smallest_enclosing_circle_toussaint_link() {
    return new SmallestEnclosingCircleToussaint();
}

Algorithm* largest_empty_circle_link() {
    return new LargestEmptyCircle();
}

