/*
 *
 * Line2D.C - Line2D & GraphicLine2D methods.
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    07.01.92  13:40 	Welson R. Jacometti  	Arquivo criado
 *
 */

#include "Obj.h"
#include "Line2D.h"
#include "utilities.h"

GraphicObj* Line2D::new_GraphicObj() {
    return new GraphicLine2D(this);
}

bool Line2D::cross_to_left( Line2D &l)  {
    if (!intersect(l))
      return false;

    Point2D p = intersection(l);
        
    // calcula t para ponto de intersecao
    double tp;
    if (l.horizontal())
      tp = (p.X() - l.x0) / l.f;
    else
      tp = (p.Y() - l.y0) / l.g;
    
    // incrementa t para gerar um ponto alem de p na direcao da reta
    tp++;
    p.setXY(l.x0 + (tp * l.f), l.y0 + (tp * l.g));
    
    return p.left_turn(p1, p2);
}

bool Line2D::cross_to_right( Line2D &l)  {
    if (!intersect(l))
      return false;

    Point2D p = intersection(l);
    
    // calcula t para ponto de intersecao
    double tp;
    if (l.horizontal())
      tp = (p.X() - l.x0) / l.f;
    else
      tp = (p.Y() - l.y0) / l.g;
    
    // incrementa t para gerar um ponto alem de p na direcao da reta
    tp++;
    p.setXY(l.x0 + (tp * l.f), l.y0 + (tp * l.g));
    
    return p.right_turn(p1, p2);
}

bool Line2D::intersect( Line2D &l)  {
    if (zero((g * l.f) - (f * l.g)))
      return false;

    return true;
}

bool Line2D::intersect( Point2D &a,  Point2D &b)  {
    Line2D l(a, b);
    double tl;

    if (!intersect(l))
      return false;

    Point2D res = intersection(l);
    if (l.horizontal())
      tl = (res.X() - l.x0) / l.f;
    else
      tl = (res.Y() - l.y0) / l.g;
      
    if (tl < 0 && a.at_infinity())
      return true;
    else
      if (tl > 1 && b.at_infinity())
        return true;

    return (tl >= 0 && tl <= 1);
}

Point2D Line2D::intersection( Line2D &l)  {
    Point2D res;
    double f1g2, f2g1, det;
    
    f1g2 = f * l.g;
    f2g1 = g * l.f;
    det = f2g1 - f1g2;
    
    if (zero(det))
      res.set_at_infinity(0, 0);	// the two lines are parallel
    else {
      double y21 = l.y0 - y0;
      double x21 = l.x0 - x0;
      double s = (l.f * y21 - l.g * x21) / det;
      double t = (f * y21 - g * x21) / det;
      res.setXY(x0 + (f * s), y0 + (g * s));
    }
    
    return res;
}

Point2D Line2D::intersection( Point2D &a,  Point2D &b)  {
    Point2D res;
        
    if (!intersect(a, b))
      res.set_at_infinity(0, 0);
    else
      res = intersection(Line2D(a, b));
    
    return res;
}

Line2D Line2D::normal_through_point( Point2D &p)  {
    Line2D res(p.X(), p.Y(), p.X() - g, p.Y() + f);
    return res;
}

IntersectType Line2D::circle_intersections( Point2D &center, double radius, Point2D &i1, Point2D &i2)  {
    double fsq, gsq, fgsq;
    double cx0, cy0;
    double fygx, fxgy, root;
    double t, fginv;
    
    fsq = f * f; gsq = g * g;
    fgsq = fsq + gsq;
    
    cx0 = center.X() - x0;
    cy0 = center.Y() - y0;
    
    fygx = (f * cy0) - (g * cx0);
    root = (radius * radius * fgsq) - (fygx * fygx);
    
    if (root < -EPSLON)
      return NoIntersection;		// reta nao intercepta circulo
    else {
      fxgy = (f * cx0) + (g * cy0);
      if (root < EPSLON) {
      	t = fxgy/fgsq;
      	i1.setXY(x0 + (f * t), y0 + (g * t));
      	return Tangent;			// reta tangente ao circulo
      }
      else {
      	root = (double)sqrt(root);
        fginv = 1.0 / fgsq;
        t = (fxgy - root) * fginv;
        i1.setXY(x0 + (f * t), y0 + (g * t));
        t = (fxgy + root) * fginv;
        i2.setXY(x0 + (f * t), y0 + (g * t));
        return Intersect;		// reta intercepta circulo
      }
    }
}

void Line2D::unfill_handles(World *w, bool show_handle)  {
    if (show_handle) {
      p1.unfill_handles(w);
      p2.unfill_handles(w);
    }
}

void Line2D::draw_handles(World *w, GC gc, bool highlight, bool show_handle)  {
    if (show_handle) {
      p1.draw_handles(w, gc, highlight);
      p2.draw_handles(w, gc, highlight);
    }
}

void Line2D::draw(World *w, GC gc)  {
    Point2D i1, i2;
    double x1, y1, x2, y2;

    w->get_values(x1, y1, x2, y2);
    if (circle_intersections(Point2D((x1 + x2)/2, (y1 + y2)/2), 2 * fabs((x1 + x2)/2 - x1) , i1, i2) == Intersect)
      w->DrawLine(gc, i1.X(), i1.Y(), i2.X(), i2.Y());
}

double Line2D::distance_to_point(double x, double y)  {
    double fsq = f * f;
    double gsq = g * g;
    double fgsq;
    if ((fgsq = fsq + gsq) < EPSLON) {	
      warning("Line2D.C", "distance_to_point", "line improperly defined");
      return TOO_FAR;
    }
    else {
      double xj0 = x - x0;
      double yj0 = y - y0;
      double fg = f * g;
      double finv = 1 / fgsq;
      double dx = (gsq * xj0) - (fg * yj0);
      double dy = (fsq * yj0) - (fg * xj0);
      return ((dx * dx) + (dy * dy)) * finv * finv;
    }
}   

void Line2D::translate_handle_at(double x, double y, double to_x, double to_y) {
    if (p1.select_test(x, y))
      p1.translate(to_x, to_y);
    else
      if (p2.select_test(x, y))
        p2.translate(to_x, to_y);

    set(p1, p2);
  
    if (p1.equalX(p2))
      subtype(Vertical_t);
    else
      if (p1.equalY(p2))
        subtype(Horizontal_t);
      else
        subtype(Generic_t);
}
        
/*
 *
 * GraphicLine2D
 *
 *
 */

// cast macro

#define line	((Line2D *)obj)

static bool is_P1;

// inherited methods redefinition

void GraphicLine2D::int_start(World *, double initial_X, double initial_Y, Int_Mode) {
    line->set(initial_X, initial_Y, initial_X, initial_Y);
}

void GraphicLine2D::int_change(World *, double new_X, double new_Y, Int_Mode) {
    line->set(line->P1().X(), line->P1().Y(), new_X, new_Y);
}

void GraphicLine2D::int_end(World *, double final_X, double final_Y, Int_Mode) {
    line->set(line->P1().X(), line->P1().Y(), final_X, final_Y);
    if (line->P1().equalX(line->P2()))
      line->subtype(Vertical_t);
    else
      if (line->P1().equalY(line->P2()))
        line->subtype(Horizontal_t);
}

void GraphicLine2D::resize_start(World *, double X, double Y, Int_Mode) {
    if (line->P1().select_test(X, Y))
      is_P1 = true;
    else
      is_P1 = false;
}

void GraphicLine2D::resize_change(World *, double X, double Y, Int_Mode) {
    if (is_P1)
      line->set(X, Y, line->P2().X(), line->P2().Y());
    else
      line->set(line->P1().X(), line->P1().Y(), X, Y);
}

void GraphicLine2D::resize_end(World *w, double X, double Y, Int_Mode m) {
    resize_change(w, X, Y, m);
    if (line->P1().equalX(line->P2()))
      line->subtype(Vertical_t);
    else
      if (line->P1().equalY(line->P2()))
        line->subtype(Horizontal_t);
      else
        line->subtype(Generic_t);
}

bool GraphicLine2D::where_draw_string (double &x, double &y) {
    Point2D p1 (((Line2D *)obj)->P1());
    Point2D p2 (((Line2D *)obj)->P2());
    x = (p1.X()+p2.X())/2;
    y = (p1.Y()+p2.Y())/2;
    return true;    
}





