/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * Obj.h - definicao da superclasse Obj, da qual descendem todos os
 *         objetos geometricos 'puros` (que nao contem dados e metodos
 *         extras, como os de interface), e da superclasse GraphicObj
 *	   que incorpora elementos do tipo Obj e prove recursos de
 *	   interface.
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    19.11.91  13:40 	Welson R. Jacometti  	Arquivo criado
 *
 */

#ifndef _OBJ_DEFINED
#define _OBJ_DEFINED

#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <iostream.h>
#include <stdio.h>
#include <sys/param.h>
#include "tools.h"
#include "World.h"
#include "String.h"
#include "utilities.h"
#include "basic.h"
#include "List.h"
#include "attributes.h"

#define TOO_FAR	10e+10

// forward delcarations
class Obj;
class GraphicObj;

typedef Obj* Obj_ptr;
typedef GraphicObj* GraphicObj_ptr;

#ifndef _Obj_ptr_list
#define _Obj_ptr_list
declare(list,Obj_ptr)
#endif ~_Obj_ptr_list;

#define GCALL 0xFFFFFFFF
     
/*****************************************************************************
  
  Para tentar diminuir o problema de nao termos a versao 3.0 do compilador
  C++, todos os metodos virtuais indefinidos em qualquer classe base sera'
  implementado de forma a produzir uma mensagem de erro do tipo:
  
  Warning: 'object type': 'method' not implemented
  
  Isso livra-me de ter que ficar contornando o problema da versao 2.0 nao
  permitir heranca sem redefinicao de metodos virtuais por mais de um ni-
  vel (somente uma classe base por hierarquia). Fique  claro  apenas que
  nestes casos, os metodos nao implementados realmente NAO tem  que  ser
  implementados, por tratar-se de uma classe abstrata.

******************************************************************************/

// As constantes abaixo referem-se ao conjunto de objetos geometricos que
// sao conhecidos (em tempo de compilacao) pelo GeoLab. Infelizmente nao
// tive tempo para pensar em um metodo mais elaborado (automatico) e seguro
// para fazer esta identificacao, tao necessaria nas funcoes que conectam
// a interface do GeoLab aos algoritmos, ligados dinamicamente.
//
// IMPORTANTE:
//
// Objetos externos ao GeoLab (aqueles que sao introduzidos pelos algoritmos
// e que nao sao necessariamente conhecidos pelo GeoLab) devem ser unicamente
// identificados para que nao ocorram conflitos entre diferentes tipos. Minha
// sugestao e' que todo novo objeto introduzido externamente seja cadastrado
// em um arquivo 'ExternalObj_id.h` de forma que fique razoavelmente facil
// dar um novo "numero" aos novos objetos.

typedef int ObjType;

const ObjType MIN_INTERNAL_OBJ_ID = 0;
const ObjType MAX_INTERNAL_OBJ_ID = 1000;

const ObjType Obj_t 		= 0;
const ObjType Point2D_t 	= 1;
const ObjType Segment2D_t	= 2;
const ObjType Ray2D_t		= 3;
const ObjType Line2D_t		= 4;
const ObjType Rectangle_t	= 5;
const ObjType Square_t		= 6;
const ObjType Circle_t		= 7;
const ObjType Ellipse_t		= 8;
const ObjType Polygon_t		= 9;

// subtypes
const ObjType Vertical_t	= 10;
const ObjType Horizontal_t	= 11;
const ObjType Generic_t		= 12;
const ObjType Simple_t		= 13;
const ObjType StarShaped_t	= 14;
const ObjType Convex_t   	= 15;
const ObjType Monotone_t	= 16;

// Funcao auxiliar, que diz se um objeto e' ou nao interno ao GeoLab

inline bool geolab_knows_object(ObjType t) {
    return range(t, MIN_INTERNAL_OBJ_ID, MAX_INTERNAL_OBJ_ID);
}

// Classe "Objeto Geometrico", raiz da hierarquia e herdado por todos seus
// componentes.

class Obj {
  public:
    // construtores 
    Obj(char *s = "Obj",  ObjType t = Generic_t, bool v = false) : is_a(s), stype(t), volati(v) {}
    
    // destrutor
    ~Obj() {}

    // construtor virtual (simulado) que estabelece o relacionamento
    //
    //                   Obj --> GraphicObj
    
    virtual GraphicObj* new_GraphicObj();

    // funcao para auto-reproducao de um objeto -> necessario quando
    // queremos, por exemplo, copiar um objeto na interface e este
    // objeto e' externo ao GeoLab.

    virtual Obj* reproduct()  {
    	warning(what_is(), "reproduct");
    	return 0;
    }
    
    // metodos para armazenamento secundario
    virtual void load(FILE *);
    virtual void save(FILE *);

    virtual ObjType type()  {
    	warning(what_is(), "type");
    	return Obj_t;
    }
    
    virtual ObjType subtype()  { return stype; }
    
    virtual ObjType subtype(ObjType v) { return stype = v; }
    
    // transformacoes geometricas aplicadas a um objeto particular
    virtual void translate(double to_x, double to_y) {
    	warning(what_is(), "translate");
    	cout << "to X: " << to_x << "   to Y: " << to_y << endl;
    }

    virtual void rotate(double orig_x, double orig_y, double angle) {
    	warning(what_is(), "rotate");
    	cout << "Origin at: " << orig_x << "," << orig_y << endl;
    	cout << "Rotation: " << angle << " radians" << endl;
    }

    virtual void scale(double factor) {
    	warning(what_is(), "scale");
    	cout << "Scale factor: " << factor << endl;
    }
    
    // metodos para descricao textual do objeto
    virtual void print()  {
    	warning(what_is(), "print");
    }

    char* what_is()  { return is_a.value(); }

    // metodos para desenho do objeto
    virtual void unfill_handles(World *, bool /*show_handle*/)  {
        warning(what_is(), "unfill_handles");
    }
    
    virtual void draw_handles(World *, GC, bool /*highlight*/, bool /*show_handle*/)  {
    	warning(what_is(), "draw_handles");
    }
    
    virtual void draw(World *, GC)  {
    	warning(what_is(), "draw");
    }

    // metodos para teste de selecao (auxilio `a interface, mantem abstracao)
    virtual bool select_test(double x, double y)  {
    	warning(what_is(), "select_test");
    	cout << "Reference point: (" << x << "," << y << ")" << endl;
    	return false;
    }

    virtual bool select_test(double x1, double y1, double x2, double y2)  {
    	warning(what_is(), "select_test");
    	cout << "Bounding box: (" << x1 << "," << y1 << ") (" << x2 << "," << y2 << ")" << endl;
    	return false;
    }

    // funcao para calculo de distancia de um ponto ao objeto, utilizado pela
    // interface para selecionar objetos, mas tambem util em varios outros casos
    virtual double distance_to_point(double x, double y)  {
    	warning(what_is(), "distance_to_point");
    	cout << "Point is at: (" << x << "," << y << ")" << endl;
    	return TOO_FAR;
    }

    // metodos para query & set de objetos volateis
    bool obj_volatile() { return volati; }
    bool obj_volatile(bool value) { return volati = value; }

    // metodos para InputTools

    virtual bool point_inside_implemented () { return false; }
    virtual bool point_inside(double x, double y)  {
        warning(what_is(), "point_inside");
        cout << "Point is at:(" << x << "," << y << ")" << endl;
        return false;
    }

    virtual bool bounding_box_implemented () { return false; }
    virtual void bounding_box (double& x1, double& y1, double& x2, double& y2)  {
        warning(what_is(), "bounding_box");
    	cout << "Bounding box: (" << x1 << "," << y1 << ") (" << x2 << "," << y2 << ")" << endl;
    }

    virtual bool get_segments_implemented () { return false; }
    virtual list(Obj_ptr) get_segments ()  {
	warning(what_is(), "get_segments");
	return 0;
    }

    // Methods for information retrieval.

     virtual bool get_points_implemented() { return false; }
     virtual list(Obj_ptr) get_points()  {
         warning(what_is(), "get_points");
         return 0;
     }

  protected:
    String is_a;
    ObjType stype;
    bool volati;
};

/*
 *
 * GraphicObj
 *
 *
 */

class GraphicObj {
  public:
    // construtores 
    GraphicObj() { obj = new Obj(); init(); }
    GraphicObj(Obj *o) { obj = o; init(); }
        
    // destrutor
    ~GraphicObj() {
    	if (gco) {			// Isto e' um tanto perigoso, mas
    	  gc_world->FreeGC(gco);	// nao da' problema enquanto nao
    	  gc_world->FreeGC(gch);	// se tem World's temporarios.
    	}
    	delete obj;
    }

    // Metodos para construcao interativa do objeto
    
    // Como a construcao se da' atraves de um modelo bem restrito de entrada
    // de dados, que sao as acoes do mouse, a criacao interativa de um objeto
    // geometrico deve seguir as fases: "int_start" para iniciar o processo de
    // criacao, onde sao passadas as coordenadas (virtuais) do primeiro click,
    // "int_change" quando o mouse e' arrastado para se acertar a posicao do
    // objeto e "int_end" quando o processo for encerrado. Para aumentar o
    // conjunto possivel de possibilidades destes metodos, existe um
    // parametro opcional (Mode) que pode vir a ser utilizado pela
    // interface para especificar diferentes maneiras de construcao
    // interativa de um mesmo objeto (e.g., circulo dado por centro e
    // raio e circulo inscrito em um quadrado).
    // A semantica do uso destas operacoes nao e' a mesma para todos os
    // objetos devido diferentes necessidades, nao sendo estabelecida pelo
    // ambiente nenhuma norma neste sentido. 
    
    virtual void int_start(World *, double X, double Y, Int_Mode) {
    	warning(what_is(), "int_start");
    	cout << "X: " << X << "   Y: " << Y << endl;
    }

    virtual void int_change(World *,double X, double Y, Int_Mode) {
    	warning(what_is(), "int_change");
    	cout << "X: " << X << "   Y: " << Y << endl;
    }

    virtual void int_end(World *,double X, double Y, Int_Mode) {
    	warning(what_is(), "int_end");
    	cout << "X: " << X << "   Y: " << Y << endl;
    }

    // Metodos para redimensionamento de objetos
    
    // Na maioria dos objetos (principalmente os objetos simples) o
    // redimensionamento se faz disparando a construcao interativa a
    // partir das dimensoes atuais do objeto (usa-se neste caso
    // dos metodos ja' construidos int_start, int_change e int_end).
    // Em outros casos porem, o redimensionamento nao repete os
    // passos da construcao interativa (por exemplo, quando nao
    // e' disponivel tal construcao) e nestes casos o redimensionamento
    // pode simplesmente afetar partes do objeto - ao gosto do implementador.
    
    virtual void resize_start(World *, double X, double Y, Int_Mode) {
    	warning(what_is(), "resize_start");
    	cout << "X: " << X << "   Y: " << Y << endl;
    }    	
    
    virtual void resize_change(World *, double X, double Y, Int_Mode) {
    	warning(what_is(), "resize_change");
    	cout << "X: " << X << "   Y: " << Y << endl;
    }    	
    
    virtual void resize_end(World *, double X, double Y, Int_Mode) {
    	warning(what_is(), "resize_end");
    	cout << "X: " << X << "   Y: " << Y << endl;
    }    	

    // metodos auxiliares        
    virtual void translate(double to_x, double to_y) {
    	obj->translate(to_x, to_y);
    }

    virtual void rotate(double orig_x, double orig_y, double angle) {
    	obj->rotate(orig_x, orig_y, angle);
    }

    virtual void scale(double factor) {
    	obj->scale(factor);
    }
    
    virtual void unfill_handles(World *w) {
    	obj->unfill_handles(w, show_handle);
    }
    
    virtual void draw_handles(World *w) {
    	obj->draw_handles(w, gch, is_selected, show_handle);
    }

    virtual void draw(World *w, DrawMode dm);

    virtual void get_string(String &s) {s = label;}
				   
    virtual void set_string(const String &s) { label = s; }    

    virtual bool where_draw_string (double &x, double &y) {
	warning(what_is(), "where_draw_string");
    	cout << "where: (" << x << "," << y << ")" << endl;
	return false;
    }	
					
    // metodos para armazenamento secundario
    virtual void load(FILE *, World *);
    virtual void save(FILE *);

    // metodos para descricao textual do objeto
    void print()  { obj->print(); }
    char* what_is()  { return is_a.value(); }

    // metodos de selecao
    virtual void select() {
    	is_selected = true;
    }
    	
    virtual void unselect() {
    	is_selected = false;
    }

    virtual bool selected() {
        return is_selected;
    }

    virtual bool select_test(double x, double y) {
    	return obj->select_test(x, y);
    }

    virtual bool select_test(double x1, double y1, double x2, double y2) {
        return obj->select_test(x1, y1, x2, y2);
    }

    virtual void glue() {
    	is_glued = true;
    }
    
    virtual void unglue() {
    	is_glued = false;
    }
    
    virtual bool glued() {
    	return is_glued;
    }
    
    virtual void anchor() {
    	is_anchored = true;
    }
    
    virtual void unanchor() { 
        is_anchored = false;
    }
    
    virtual bool anchored() {
    	return is_anchored;
    }
    
    virtual double distance_to_point(double x, double y) {
	return obj->distance_to_point(x, y);
    }

    // metodos para manipula,c~ao de atributos
    virtual void set_attributes(World *w, Attributes& attr);

    virtual Attributes get_attributes();

    // metodos para manipulacao do GC do objeto (auxiliares a varias outras)
    virtual void set_handle(bool sh) {
    	show_handle = sh;
    }
    
    bool handle_is_on() {
    	return show_handle;
    }
    
    virtual void set_color(World *w, unsigned long fg, unsigned long bg) {
    	gc_init(w);
    	w->SetForeground(gco, fg); w->SetBackground(gco, bg);
    	w->SetForeground(gch, fg); w->SetBackground(gch, bg);
    	foreground = fg; background = bg;
    }

    virtual void get_color(unsigned long &fg, unsigned long &bg) {
    	fg = foreground; bg = background;
    }

    virtual void set_line_attr(World *w, int lw, int ls) {
    	gc_init(w);
    	w->SetLineAttributes(gco, lw, ls, 0, 0);	// not gch
    	line_width = lw; line_style = ls;
    }
    
    virtual void get_line_attr(int &lw, int &ls) {
    	lw = line_width; ls = line_style;
    }
    
    virtual void set_all(World *w, unsigned long fg, unsigned long bg, int lw, int ls, bool sh) {
        set_color(w, fg, bg);
        set_line_attr(w, lw, ls);
        set_handle(sh);
    }

    virtual GC get_gco() { return gco; }
    virtual GC get_gch() { return gch; }
    
    virtual void set_draw_mode(World *w, DrawMode dm) {
    	gc_init(w);
        
        if (dm == CopyMode) {
          w->SetFunction(gco, GXcopy);
          w->SetFunction(gch, GXcopy);
        }
        else {
    	  w->SetFunction(gco, GXxor);
    	  w->SetFunction(gch, GXxor);
    	}
    }

    // funcao para auto-reproducao de um objeto -> necessario quando
    // queremos, por exemplo, copiar um objeto na interface e este
    // objeto e' externo ao GeoLab.

    virtual GraphicObj* reproduct() {
    	Obj *reproducted_obj = obj->reproduct();
    	if (reproducted_obj) {
      	  GraphicObj *go = reproducted_obj->new_GraphicObj();
      	  go->ext_function = ext_function;
      	  go->ext_module = ext_module;
          go->label = label;
    	  go->show_handle = show_handle;
          go->foreground = foreground;
          go->background = background;
          go->line_width = line_width;
          go->line_style = line_style;
          if (gco) {
            go->gc_world = gc_world;
            go->gco = gc_world->CreateGC(0, NULL);
            gc_world->CopyGC(gco, GCALL, go->gco);
            go->gch = gc_world->CreateGC(0, NULL);
            gc_world->CopyGC(gch, GCALL, go->gch);
          }
          return go;
        }
        else
          return 0;
    }

    // suporte para identificacao da origem (onde foi criado) de objetos
    // externos, util para recuperacao de objetos de memoria secundaria

    ObjType type() { return obj->type(); }
    
    ObjType subtype() { return obj->subtype(); }
    ObjType subtype(ObjType v) { return obj->subtype(v); }
    
    void set_external(char *function, char *module) {
        ext_function = function;
        ext_module = module;
    }

    char* get_external_function()  { return ext_function.value(); }

    char* get_external_module()  { return ext_module.value(); }
        
    // suporte para objetos volateis
            
    bool obj_volatile() { return obj->obj_volatile(); }
    bool obj_volatile(bool value) { return obj->obj_volatile(value); }
    
    Obj *obj;
    
  protected:
    void init() {
    	is_a = String("Graphic") + String(obj->what_is());
        show_handle = !Default_Obj_Attributes.HANDLES;
	is_glued = !Default_Obj_Attributes.GLUE;
	is_anchored = Default_Obj_Attributes.DELETABLE;
	foreground = Default_Obj_Attributes.FOREGROUND;
	background = Default_Obj_Attributes.BACKGROUND;
	line_width = Default_Obj_Attributes.LINE_WIDTH;
	line_style = Default_Obj_Attributes.LINE_STYLE;
	is_selected = false;	
	saved_dm = NoMode;
	gch = gco = 0;
	gc_world = 0;
    }
    
    void gc_init(World *w) {
    	if (!gco) {
    	  gc_world = w;
    	  gco = w->CreateGC(0, NULL);
    	  gch = w->CreateGC(0, NULL);
    	}
    }

    String is_a;
    String ext_function, ext_module;
    String label;
    bool show_handle;
    bool is_selected;
    bool is_glued, is_anchored;
    DrawMode saved_dm;
    unsigned long foreground, background;
    int line_width, line_style;
    GC gco, gch;
    World *gc_world;		// destrutor precisa para poder desalocar GC's
};

#endif ~_OBJ_DEFINED;









