/*
    Copyright (C) 1992-1994
                    Laerte F. Morgado

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 *            Pslg_geotools.C: Modulo que contem a implementacao de algumas
 *                           ferramentas geometricas basicas.
 *
 *
 *            Laerte F. Morgado          01/03/91    arquivo criado.
 *
 *
 *
 */


#include <iostream.h>
#include "tools.h"
#include "Pslg_geotools.h"


#define EPS 1e-10


/*
 *  "planar_cross_product":
 *
 *    Returns the value of the z-coordinate of the cross product
 *  between the two vectors whose endpoints are passed as
 *  parameters. The first coordinates refer to the first
 *  vector in the product.
 */


double planar_cross_product(double x1,double y1,double x2,double y2,double EPSON)
{
    double r;
    r = (x1*y2-y1*x2);
    if (zero(r,EPSON)) r=0;
    return r;
}


double planar_dot_product(double x1,double y1,double x2,double y2,double EPSON)
{
    double r;
    r = x1*x2 + y1*y2;
    if (zero(r,EPSON)) r=0;
    return r;
}
     


/*
 *  "left_side":
 *
 *    Returns "true" if the point (x,y) is on the left side of the 
 *  oriented angle (x1,y1)->(x2,y2)->(x3,y3) whose corner is (x2,y2).
 *
 */
 
     
bool  left_side(double x,double y,double x1, double y1,double x2,double y2,double x3,double y3)
{
    double p1,p2,p;
    p1 = planar_cross_product(x1-x,y1-y,x2-x1,y2-y1);
    p2 = planar_cross_product(x2-x,y2-y,x3-x2,y3-y2);
    p  = planar_cross_product(x2-x1,y2-y1,x3-x2,y3-y2);

    if (p>0)
      return (p1>=0 && p2>=0);
    else if (p<0)
      return (!((p1<0)&&(p2<0)));
    else
      {
	  if (planar_dot_product(x2-x1,y2-y1,x3-x2,y3-y2)>0)
	    return (p1>=0 && p2>=0);
	  else
	    return (!((p1<0)&&(p2<0)));
      }
}




/*
 *  "intersect":
 *
 *      Determines if the two planar segments (x1,y1)-(x2,y2) and 
 *   (x3,y3)-(x4,y4) intersect. The parameter mode1 allows to 
 *   specify the status of the endpoints of the first sement
 *   among four possibilities: both closed, both open, first open and 
 *   second open. The parameter mode2 is the similar one to the second
 *   segment. Returns "true" if the two segments intersect and "false"
 *   otherwise. Weird cases are treated.
 */




      
bool intersect( double x1,double y1, double x2,double y2, double x3,double y3, double x4, double y4,InterType mode1,InterType mode2)
{
   bool inter=true;      // True if the two segments intersect.
   double A,B,C,D,E,F;
     //  A.alpha1 + B.alpha2 = C;   (I)
     //  D.alpha1 + E.alpha2 = F;
     //    The above equations system is the result of the following one:
     //
     //  alpha1.(x1,y1) + beta1.(x2,y2) = alpha2.(x3,y3) + beta2.(x4,y4) (III)
     //       where
     //            alpha1+beta1=1;
     //            alpha2+beta2=1;      (II)
     //            0<=alpha1<=1;
     //            0<=alpha2<=1;
     //
     //   Explaining:
     //   Every point in a segment is a linear combination of its two extrems
     // satisfying conditions (II). If two segments intersect, then
     // one point of one must be equal one point of another. This explains
     // equation (III).
     
   double alpha1,beta1,alpha2,beta2;
   double delta,alpha1_delta,alpha2_delta;
   
     //   Variable delta stores the value of the derterminant of the
     /* matrix  |A B|  when solving system I by Cramer's rule   */
     /*         |D E| .                                         */
     //   Variable alpha1_delta stores the value of the
     // determinant | C B |.
     //             | F E |
     //   Variable alpha2_delta stores the value of the
     // determinant | A C |.
     //             | D F |


   A = x2-x1; B = x3 - x4; C = x2 - x4;
   D = y2-y1; E = y3 - y4; F = y2 - y4;
   delta = A*E - B*D;
   alpha1_delta = C*E - B*F;
   alpha2_delta = A*F - C*D;

   
   if (!zero(delta,EPS))   // system (I) is solvable and has just one solution.
     {

	 alpha1 = alpha1_delta/delta;  /* Cramer's rule. */
	 alpha2 = alpha2_delta/delta;  /* Cramer's rule. */
	 
	 beta1 = 1-alpha1;             // Obeying restrictions (I).
	 beta2 = 1-alpha2;             // Obeying restrictions (I).
	 if ((less(alpha1,0,EPS))||(less(1,alpha1,EPS))||(less(alpha2,0,EPS))||(less(1,alpha2,EPS)))
             inter=false;
         else
	   {
	       if ((zero(alpha1,EPS))&&( (mode1==SECOND_OPEN)||(mode1==BOTH_OPEN)) )
		 // Intersection is at point (x2,y2).
		   inter=false;
	       else
		 if ((zero(beta1,EPS))&& ( (mode1==FIRST_OPEN) ||(mode1==BOTH_OPEN)) )
		 // Intersection is at point (x1,y1).
		   inter=false;	 
	       else
		 if ((zero(alpha2,EPS))&&( (mode2==SECOND_OPEN)||(mode2==BOTH_OPEN)) )
		 // Intersection is at point (x4,y4).
		   inter=false;
	       else
		 if ((zero(beta2,EPS))&& ( (mode2==FIRST_OPEN) ||(mode2==BOTH_OPEN)) )
		 // Intersection is at point (x3,y3).
		   inter=false;
	   }
     }
   else   // System (I) is unsolvable or has infinit solutions.
     {
	 
	 if ((!zero(alpha1_delta,EPS))||(!zero(alpha2_delta,EPS)))
	   // System (I) is unsolvable.
	   inter=false;
	 else  // System (I) has infinit solutions .
	       /* Let's see if they can satisfy the restrictions. */
	   {
	       double U,V,W; 
	       if ((zero(A,EPS))&&(zero(B,EPS))&&(zero(C,EPS)))
		 {
		     U=D;V=E;W=F;
		 }
	       else
		 {
		     U=A;V=B;W=C;
		 };
               //   The two equations os system (I) are proportional.
	       // Just rejected the one with 0 coeficients.
	       // U, V and W are the coeficients of the other one.
	       
	       
	       if ((zero(U,EPS))&&(zero(V,EPS)))
		 if (!zero(W,EPS))
		       inter=false;
		 else  //  (x1,y1)=(x2,y2)=(x3,y3)=(x4,y4).
		   if ((mode1==BOTH_OPEN)||(mode2==BOTH_OPEN))
		     // One segment has no points at all.
		     inter=false;
	       
	       if ((zero(U,EPS))&&(!zero(V,EPS)))
		 // (x1,y1)=(x2,y2) e (x3,y3)!=(x4,y4).
		 {
		     if (mode1==BOTH_OPEN) inter=false;
		     else
		       {
			   alpha2=W/V;
			   if ((less(alpha2,0,EPS))||(less(1,alpha2,EPS)))
			     inter = false;
			   else if
			     ((zero(alpha2,EPS))&&((mode2==SECOND_OPEN)||(mode2==BOTH_OPEN)))
			     //(x1,y1)=(x2,y2)=(x4,y4)!=(x3,y3).
			     inter =false;
			   else if
			     ((equal(alpha2,1,EPS))&&((mode2==FIRST_OPEN)||(mode2==BOTH_OPEN)))
			     // (x1,y1)=(x2,y2)=(x3,y3)!=(x4,y4).
			     inter=false;
		       }
		 };
	       
	       if ((!zero(U,EPS))&&(zero(V,EPS)))
		 // (x1,y1)!=(x2,y2) e (x3,y3)=(x4,y4);
		 {
		     if (mode2==BOTH_OPEN) inter=false;
		     else
		       {
			   alpha1=W/U;
			   if ((less(alpha1,0,EPS))||(less(1,alpha1,EPS)))
			     inter = false;
			   else if
			     ((zero(alpha1,EPS))&&((mode1==SECOND_OPEN)||(mode1==BOTH_OPEN)))
                             // (x1,y1)!=(x2,y2)=(x3,y3)=(x4,y4).
			     inter =false;
			   else if
			     ((equal(alpha1,1,EPS))&&((mode1==FIRST_OPEN)||(mode1==BOTH_OPEN)))
			     // (x2,y2)!=(x1,y1)=(x3,y3)=(x4,y4).
			     inter=false;
		       }	   
		     
		 };
	       
	       if ((!zero(U,EPS))&&(!zero(V,EPS)))
		 // (x1,y1)!=(x2,y2) e (x3,y3)!=(x4,y4).

		 {
		     double inf,sup;
		     // Here, we force 0<=alpha1<=1.
		     // The we obtain the limits os alpha2.
		     if (less(0,V/U,EPS))
		       {
			   inf= (W/U -1)*U/V;  // inferior limit.
			   sup= W/V;           // superior limit.
		       }
		     else
		       {
			   sup= (W/U -1)*U/V;  // inferior limit.
			   inf= W/V;           // superior limit.
		       };

		     if ((less(sup,inf,EPS))||(less(sup,0,EPS))||(less(1,inf,EPS)))
		       inter=false;
		     else if (zero(sup,EPS))
		       {
			   if ((zero(W,EPS))&&((mode1==SECOND_OPEN)||(mode1==BOTH_OPEN)||(mode2==SECOND_OPEN)||(mode2==BOTH_OPEN)))
			     // (x2,y2)==(x4,y4).
			     inter=false;
			   else if ((equal(W,U,EPS))&&((mode1==FIRST_OPEN)||(mode1==BOTH_OPEN)||(mode2==SECOND_OPEN)||(mode2==BOTH_OPEN)))
			     // (x1,y1)==(x4,y4)
			     inter=false;

		       }
		     else if (equal(inf,1,EPS))
		       {
			   if ((equal(W,V,EPS))&&((mode1==SECOND_OPEN)||(mode1==BOTH_OPEN)||(mode2==FIRST_OPEN)||(mode2==BOTH_OPEN)))
			     // (x2,y2)=(x3,y3);
			     inter=false;
			   else if ((equal(W,U+V,EPS))&&((mode1==FIRST_OPEN)||(mode1==BOTH_OPEN)||(mode2==FIRST_OPEN)||(mode2==BOTH_OPEN)))
			     // (x1,y1)=(x3,y3);
			     inter=false;
			   
		       }
		     
		      
		 } 
	       

	   }
	 
     };
   return inter;
}   



/*
 * Calcula o "angulo" anti-horario que o vetor (x1,y1)->(x2,y2)
 * faz com o eixo X.
 * 
 * (0,0)->(1,0) faz 0;
 * (0,0)->(0,1) faz 90;
 * (0,0)->(-1,0) faz 180;
 * (0,0)->(0,-1) faz 270;
 */

double Theta(double x1, double y1, double x2, double y2)
{
    double dx,dy,ax,ay,t;
    
    ax = fabs(dx = x2 - x1);
    ay = fabs(dy = y2 - y1);
    if ( (dx == 0)&&(dy == 0) )
      t=0;
    else t = dy/(ax+ay);
    if (dx < 0) t = 2-t;
    else if (dy < 0) t=4+t;
    return (t*90.0); 
}


