/*
    Copyright (C) 1992-1994
                    Laerte F. Morgado

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 *   Pslg_tools.C:  Contem a impementacao de ferramentas que trabalham
 *                  com a estrutura Pslg. Em particular, com a estrutura
 *                  de dados "half-edge" de suporte.
 *
 *
 *   Laerte Ferreira Morgado    01/03/92        arquivo criado.
 *
 */

#include <time.h> // for "Unreasonable include nesting"
#include "Pslg_tools.h"
#include "Line2D.h"
#include "Obj.h"
#include "External.h"


static Pslg_tools_inter_ptr First_Pslg_tools_inter_ptr;


/*
 *  "vertex_gravity" :   Determina o vertice "vertex" do PSLG "solid" 
 *                     que esta' mais proximo do ponto (x,y) e a uma
 *                     distancia minima de "VERTEX_GRAVITY" desse ponto.
 *                     As coordenadas desse vertice sao retornadas nos
 *                     mesmos parametros de entrada x e y. Devolve 
 *                     "true" se tal vertice foi encontrado e "false"
 *                     no caso contrario. O vertice, se encontado, e'
 *                     devolvido no parametro "vertex".
 *
 */

int vertex_gravity(Solid* solid,double& x,double& y,Vertex*& vertex,double X_gravity, double Y_gravity)
{
    double D1=0,D2=0,x1,y1;
    Vertex* vert;
    vertex=0;
    if ( (vert = solid->sverts) != 0 )
      {
	  x1 = vert->vcoord[Space_XCoord];y1=vert->vcoord[Space_YCoord];
	  D1 = D2 = max(fabs(x1-x),fabs(y1-y));
	  vertex  = vert;
	  vert    = vert->nextv;
	  while (vert!=0)
	    {
		x1 = vert->vcoord[Space_XCoord];y1=vert->vcoord[Space_YCoord];
		D2 = max(fabs(x1-x),fabs(y1-y)); 
		if (D2 < D1)
		  {
		      D1 = D2;
		      vertex = vert;
		  };
		vert = vert->nextv;
	    };
      };
    if ((vertex!=0) && (fabs(vertex->vcoord[Space_XCoord]-x)<X_gravity) && (fabs(vertex->vcoord[Space_YCoord]-y)<Y_gravity)) 
      {
	  x = vertex->vcoord[Space_XCoord];
	  y = vertex->vcoord[Space_YCoord];
	  return true;
      }
    else
      {
	  vertex=0;
	  return false;
      }
}



/*
 *  "intersect":   Determina o numero de intersecoes de um dado
 *               segmento com os segmento do solido "solid".
 *               O parametro mode1 especifica o estado das extremidades
 *               dos segmentos do solido (ambas abertas, ambas fechadas,
 *               primeira aberta ou segunda aberta. O parametro mode2
 *               especifica o estado das extremidades do segmento em 
 *               questao.
 */
  

long intersect(Solid* solid, double x1,double y1,double x2,double y2,InterType mode1,InterType mode2)
{
    Edge* edge; Vertex *aux1=0,*aux2=0;long counter=0;
    edge = solid->sedges;
    if (edge==0) return counter;
    do
      {
	  aux1=edge->he1->vtx;
	  aux2=edge->he1->nxt->vtx;
	  if (intersect(aux1->vcoord[Space_XCoord],aux1->vcoord[Space_YCoord],aux2->vcoord[Space_XCoord],aux2->vcoord[Space_YCoord],x1,y1,x2,y2,mode1,mode2))
	    counter++;
	  edge = edge->nexte;
      }
    while (edge!=0);
    return counter;
}
	



/*
 * "find_face" :
 *      
 *     Finds the face determined by two of the edges emanating 
 *  from vertex "vertex" in which point (x,y) is located.
 *
 */



void find_face(Face*& face, Vertex* vertex, double x, double y)
{
    HalfEdge *hf1, *hf2 , *safe;

    if (vertex==0) {face=0;return;}
    if ((safe=hf1=vertex->vedge)==0) return;
    hf2=mate(hf1);
    if (hf2==0)  {face=hf1->wloop->lface;return;}  // Nao funciona !!!
    else
      do
	{
	    hf2 = hf1->prv;
	    if (left_side(x,y,hf2->vtx->vcoord[Space_XCoord],hf2->vtx->vcoord[Space_YCoord],hf1->vtx->vcoord[Space_XCoord],hf1->vtx->vcoord[Space_YCoord],hf1->nxt->vtx->vcoord[Space_XCoord],hf1->nxt->vtx->vcoord[Space_YCoord]))
	      {
		  face = hf1->wloop->lface;
		  break;
	      };
	    hf1 = mate(hf2);
	}
      while (hf1!=safe);
}
    


void Pslg_intersect(Pslg_tools_inter& it,HalfEdge*& e, Segment2D& s)
{
    static double NEIGHBOURHOOD_FACTOR = 0.8;
    double x1,y1,x2,y2;
    Segment2D se;
    x1 = e->X();
    y1 = e->Y();
    x2 = e->nxt->X();
    y2 = e->nxt->Y();

    se.set(x1,y1,x2,y2);
    it.status = NO_CR;  /* dont intersect.*/
    it.vertex = 0;      /* no vertex.     */
    it.he     = e;


    if (s.intersect(se))
      {
	  it.p = s.intersection(se);
	  it.status = ONE_IN_CR;     // intersect in one point 
	  if (it.p.at_infinity())    // coincident lines.
	    it.status = MANY_CR;     // infinit intersections
      };

    if (it.status != MANY_CR)
      {
	  if (zero(s.distance_to_point(x1,y1),VERTEX_GRAVITY*NEIGHBOURHOOD_FACTOR))
	    {
		it.vertex = e->vtx;
		it.p.setXY(x1,y1);
		it.status = ONE_VERTEX_CR;     // intersect in one point 
	     }
	  else if (zero(s.distance_to_point(x2,y2),VERTEX_GRAVITY*NEIGHBOURHOOD_FACTOR))
	    {
		it.vertex = e->nxt->vtx;
		it.p.setXY(x2,y2);
		it.status = ONE_VERTEX_CR;    // intersect in one point 
	     };   
	  
      }

}





static int Pslg_inter_cmp(Pslg_tools_inter_ptr& a, Pslg_tools_inter_ptr& b)
{
    double d1,d2;
    
    if ((d1 = a->p.e_distance( First_Pslg_tools_inter_ptr->p)) <  (d2 = b->p.e_distance(First_Pslg_tools_inter_ptr->p)))
      return -1;
    else if (d2 < d1) return 1;
    else return 0;
}


void kill_Pslg_tools_list(list(Pslg_tools_inter_ptr)& l)
{
    list_item it,ni;
    if (!l.empty())
      {
	  it = l.first();
	  ni = l.succ(it);
	  
	  while (ni!=0)
	    {
		delete l[it];
		l.del_item(it);
		it = ni;
		ni = l.succ(it);
	    };
	  delete l[it];
	  l.del_item(it);
      }
}


void show_Pslg_tools_list(list(Pslg_tools_inter_ptr)& l)
{
    list_item it;
    Pslg_tools_inter_ptr E;
    int i=0;
    if (!l.empty())
      {
	  it = l.first();
	  while (it!=0)
	    {
		i++;
		E = l.inf(it);
		cerr<<"    Element: "<<i<<endl;
		if (E->vertex != 0)
		  cerr<<"            Vertex is not null: "<<endl;
		else
		  cerr<<"            Vertex is null: "<<endl;
		if (E->he != 0)
		  cerr<<"            HE is not null: "<<endl;
		else
		  cerr<<"            HE is null: "<<endl;
		cerr<<"           ( "<<E->p.X()<<" , "<<E->p.Y()<<" )"<<endl;
		cerr<<endl;
		it = l.succ(it);
	    };
	  cerr<<" Number of Elements is "<<i<<endl;
      }
}



void copy_Pslg_tools_list(list(Pslg_tools_inter_ptr)& old_list,list(Pslg_tools_inter_ptr)& new_list)
{
    while (!old_list.empty())
      new_list.append(old_list.pop());
}


bool All_Pslg_intersect(list(Pslg_tools_inter_ptr)& l, Solid*& solid,Segment2D& s)
{
   Edge* edge;
   Pslg_tools_inter_ptr aux;
   list_item it,ni;


   if (zero(s.P1().e_distance(s.P2()),VERTEX_GRAVITY)) return false;
   
   First_Pslg_tools_inter_ptr = aux = new Pslg_tools_inter;
   aux->p = s.P1();
   aux->vertex = 0;
   aux->he = 0;
   l.push(aux);


   aux = new Pslg_tools_inter;
   aux->p = s.P2();
   aux->vertex = 0;
   aux->he = 0;
   l.append(aux);
   
   if (solid == 0)
     return true;

   edge = solid->sedges;

   while (edge)
     {
	 aux = new Pslg_tools_inter;
	 Pslg_intersect(*aux, edge->he1,s);
	 if (aux->status == ONE_IN_CR || aux->status == ONE_VERTEX_CR)
	   l.append(aux);
	 else if (aux->status == MANY_CR)
	   {
	       kill_Pslg_tools_list(l);
	       return false;
	   };
	 edge = edge->nexte;
     };




   l.sort(Pslg_inter_cmp);

   
   it = l.first();
   ni = l.succ(it);

   while ( it && ni)
     {
	 if ((l.inf(it)->p == l.inf(ni)->p) || (zero(l.inf(it)->p.e_distance(l.inf(ni)->p),VERTEX_GRAVITY)))
	   {

	       if (l.inf(it)->vertex == 0)
		 {
		     delete l[it];
		     l.del_item(it);
		     it = ni;
		 }
	       else
		 {
		     delete l[ni];
		     l.del_item(ni);
		 }
	   }
	 else it = ni;
	 ni = l.succ(it);
     };

   ni = l.last();
   s.setP2(l.inf(ni)->p.X(),l.inf(ni)->p.Y());

   return true;
   
}


static void Pslg_low_apply_Euler(Pslg& pslg,Id solid_id,Pslg_tools_inter_ptr& a,Pslg_tools_inter_ptr& b,Id& max_vertex, Id& max_face, bool flying )
{
    Face* face;
    Vertex* v;

    if (flying)
      {
	  v = a->vertex;
	  find_face(face,v,(a->p.X()+b->p.X())/2,(a->p.Y()+b->p.Y())/2);
	  pslg.geo_smev(solid_id,face->faceno,v->vno,++max_vertex,b->p.X(),b->p.Y());
      }
    else
      {
	  if (b->vertex == 0)
	    {
		lmev(b->he,mate(b->he)->nxt,++max_vertex,b->p.X(),b->p.Y(),0);
		b->vertex = pslg.getVertex(solid_id,max_vertex);
	    };
	  v = b->vertex;
	  find_face(face,v,(a->p.X()+b->p.X())/2,(a->p.Y()+b->p.Y())/2);
	  pslg.getFirsts()->getmaxnames(max_face,max_vertex);
	  pslg.geo_mef(solid_id,face->faceno,a->vertex->vno,b->vertex->vno,++max_face);
      }
}	  
	  
	  
	  


bool Pslg_apply_Euler(Pslg& pslg,Id solid_id,list(Pslg_tools_inter_ptr)& l)
{
    Solid* solid;
    list_item it,ni;
    Id max_face, max_vertex;

    if (l.empty()) return true;
   
    solid = pslg.getSolid(solid_id);
    
    it = l.first(); ni = l.succ(it);
    if (solid == 0)
      {
	  pslg.mvfs(solid,solid_id,max_face=0,max_vertex=0,l.inf(it)->p.X(),l.inf(it)->p.Y());
	  l.inf(it)->vertex = pslg.getVertex(solid_id,max_vertex);
      }
    else
      solid->getmaxnames(max_face,max_vertex);

    if (l.inf(it)->vertex == 0)
      {
	  cerr << " Not allowed"<<endl;
	  return false;
      };

   while (ni !=0)
     {
	Pslg_low_apply_Euler(pslg,solid_id,l.inf(it),l.inf(ni),max_vertex,max_face,((ni==l.last())&& (l.inf(ni)->vertex == 0)));
	it = ni;
	ni = l.succ(it);
     };
    
    return true;

}
     
	 
void Pslg_apply_graphical_Euler(list(Pslg_tools_inter_ptr)& l,World* world,DrawMode dm)
{
    list_item it,ni;
    GraphicPoint2D   p;
    GraphicSegment2D  s;


    s.set_color(world,FOREGROUND,BACKGROUND);
    s.set_line_attr(world,LINE_WIDTH,LINE_STYLE);

    p.set_color(world,FOREGROUND,BACKGROUND);
    p.set_line_attr(world,LINE_WIDTH,LINE_STYLE);
    
    it = l.first(); ni = l.succ(it);

    while (ni !=0)
      {
	 // if (l.inf(it)->vertex == 0)
	   // {
		//cerr<<" drawing "<<endl;

		//*((Point2D*)(p.obj)) = l.inf(it)->p;
		//p.draw(world,XorMode);
		
	    //};
	  (*((Segment2D*)(s.obj))).set(l.inf(it)->p,l.inf(ni)->p);
	  s.draw(world,dm);
	  it = ni;
	  ni = l.succ(it);
	  
      };
    //if (l.inf(it)->vertex == 0)
      //{
	//  *((Point2D*)(p.obj)) = l.inf(it)->p;
	  //p.draw(world,XorMode);
      //}
}





static double X(HalfEdge* he)
{
    return (he->vtx->vcoord[Space_XCoord]);
}


static double X(Vertex* v)
{
    return v->vcoord[Space_XCoord];
}


static double Y(Vertex* v)
{
    return v->vcoord[Space_YCoord];
}

static double Y(HalfEdge* he)
{
    return (he->vtx->vcoord[Space_YCoord]);
}


static HalfEdge* next(HalfEdge* he,Vertex* v)
{
    if (he->nxt->nxt->vtx == v)
      return he->nxt->nxt;
    else
      return he->nxt;
}


static HalfEdge* prev(HalfEdge* he,Vertex* v)
{
    if (he->vtx == v)
      return he->prv->prv;
    else
      return he->prv;
}


void get_face_polygon(Vertex* vi,Vertex* vf,HalfEdge*& he1,HalfEdge*& he2,Face* face)
{
    HalfEdge *the_mate;
    Vertex* v;
    he1 = he2 = face->findhe1(vi->vno);
    while(he2->vtx!=vf)
      {
	  he2 = he2->nxt;
	  if (he2->vtx == vi) he1 = he2;
      };

    v = he1->nxt->vtx;
    the_mate = mate(he1);
    he1 = prev(he1,vi);
    while ((he1->nxt!=the_mate)&&left_side(X(he1->nxt),Y(he1->nxt),X(vf),Y(vf),X(vi),Y(vi),X(v),Y(v)))
      {
	  he1 = prev(he1,vi);
      };
    he1 = next(he1,vi);

    v = he2->prv->vtx;
    the_mate = mate(he2->prv);
    while ((he2!=the_mate)&&left_side(X(he2->nxt),Y(he2->nxt),X(v),Y(v),X(vf),Y(vf),X(vi),Y(vi)))
      {
	  he2 = next(he2,vf);
      };

}




/*
 * Angle(): retorna o "angulo" que o vetor de origem "vet" e
 *          com destino o outro vertice de "edge" faz com o
 *          eixo X.
 */

double Angle(HalfEdge_ptr edge,Vertex_ptr vet)
{
    if (edge->vtx == vet)
      return Theta(vet->X(),vet->Y(),edge->nxt->X(),edge->nxt->Y());
    else
      return Theta(vet->X(),vet->Y(),edge->X(),edge->Y());
}




/*
 * FirstIncoming(): devolve a primeira HalfEdge do terceiro
 *                  quadrante no precurso anti-horario
 *                  das arestas que circundam o vertice vet.
 *                  O vertice vet e' o destino da aresta encontrada.
 */

HalfEdge_ptr FirstInComing(Vertex_ptr vet)
{
    HalfEdge_ptr aux,aux2;
    double  ang1,ang2;

    aux2 = aux = vet->vedge->prv;
    if (Angle(aux,vet) <= 180)
      {
	  do
	    {
		aux = mate(aux)->prv;
	    }
	  while ( (aux != aux2)&&(Angle(aux,vet) <= 180) );
	  if (aux == aux2) return 0;
	  else return aux;
      }
    else 
      {
	  aux2 = aux;
	  ang2 = Angle(aux2,vet);
	  do
	    {
		aux  = aux2; ang1 = ang2;
		aux2 = mate(aux->nxt);
		ang2 = Angle(aux2,vet);
	    } 
	  while ((ang2 < ang1)&&(ang2 > 180));
	  return aux;
      }
}



/*
 * Incomnig(): DEVE SER usado apos o FirstIncoming(). Devolve
 *             a proxima HalfEdge no sentido anti-horario a
 *             partir da aresta fornecida como parametro.
 *             A aresta devolvida tem como seu vertice destino
 *             o vertice ao redor do qual esta sendo feito
 *             o percurso. Quando a HalfEdge a ser devolvida for
 *             a HalfEdge original do percurso ou quando estiver
 *             no primeiro ou segundo quadrantes, 'e devolvido 0.
 */

HalfEdge_ptr InComing(Vertex_ptr vet,HalfEdge_ptr edge)
{
    double ang1,ang2;
    if (edge == 0) return 0;
    else
      {
	  ang1 = Angle(edge,vet);
	  edge = mate(edge)->prv;
	  ang2 = Angle(edge,vet);
	  if (ang1 < ang2) return edge;
	  else return 0;
      }
} 



/*
 * FirstOutcoming(): devolve a primeira HalfEdge do segundo
 *                  quadrante no precurso horario
 *                  das arestas que circundam o vertice vet.
 *                  O vertice vet e' a origem da aresta encontrada.
 */

HalfEdge_ptr FirstOutComing(Vertex_ptr vet)
{
    HalfEdge_ptr aux,aux2;
    double ang1,ang2;
    
    aux2 = aux = vet->vedge;
    if (Angle(aux,vet) >= 180)
      {
	 do
	   aux = mate(aux)->nxt;
	 while ((aux != aux2)&&(Angle(aux,vet) >= 180));
	 if ((aux != aux2)&&(Angle(aux,vet) > 0)) return aux;
	 else return 0;
     }
    else
      {
	  aux2 = aux;
	  ang2 = Angle(aux2,vet);
	  do
	    {
		aux  = aux2;
		ang1 = ang2;
		aux2 = mate(aux2->prv);
		ang2 = Angle(aux2,vet);
	    } 
	  while ((ang2 > ang1)&&(ang2 < 180));
	  if (ang1 > 0) return aux;
	  else return 0;
      }
}





/*
 * Outcoming(): DEVE SER usado apos o FirstOutcoming(). Devolve
 *             a proxima HalfEdge no sentido horario a
 *             partir da aresta fornecida como parametro.
 *             A aresta devolvida tem como seu vertice origem
 *             o vertice ao redor do qual esta sendo feito
 *             o percurso. Quando a HalfEdge a ser devolvida for
 *             a HalfEdge original do percurso ou quando estiver
 *             no terceiro ou quarto quadrantes, e' devolvido 0.
 */


HalfEdge_ptr OutComing(Vertex_ptr vet,HalfEdge_ptr edge)
{
    double ang1,ang2;
    if (edge == 0) return 0;
    else
      {
	  ang1 = Angle(edge,vet);
	  edge = mate(edge)->nxt;
	  ang2 = Angle(edge,vet);
	  if ((ang1 > ang2)&&(ang2 > 0)) return edge;
	  else return 0;
      }
}


/*
 * find_horiz():
 *   
 *     Devolve as HalfEdges horizontais que partem de um vertice v, ambas
 *  com origem em v. A HE left dirige-se para a esquerda e a HE right
 *  dirige-se para a direita. Se nao existirem sao devolvidas como 0.
 */

void find_horiz(Vertex_ptr v,HalfEdge_ptr& left,HalfEdge_ptr& right)
{
    HalfEdge_ptr first,runner;

    left = right = 0;
    runner = first = v->vedge;
    do 
      {
	  //cerr<<"laerte3  "<<runner->X()<<endl;
	  //cerr<<runner->Y()<<endl;
	  //cerr<<v->X()<<endl;
	  //cerr<<v->Y()<<endl;
	  if ((runner->nxt->Y() == v->Y()) && (runner->nxt->X() < v->X()))
	    {
		left = runner;
		//cerr<<"laerte1  "<<runner->Y()<<endl;
		//cerr<<v->Y()<<endl;
	    }
	  if ((runner->nxt->Y() == v->Y()) && (runner->nxt->X() > v->X()))
	    {
		right = runner;
		//cerr<<"laerte2  "<<runner->Y()<<endl;
		//cerr<<v->Y()<<endl;
	    }
	  //cerr<<"laerte3  "<<runner->Y()<<endl;
	  runner = mate(runner)->nxt;
      } while (runner != first);
}

void find_min_max_x_y( Pslg& pslg,Vertex_ptr& minx,Vertex_ptr& miny,Vertex_ptr& maxx,Vertex_ptr& maxy)
{
    Vertex_ptr v;
    v = minx = miny = maxx = maxy = pslg.getFirsts()->sverts;
    v = v->nextv;
    while (v != 0)
      {
	  if (v->X() < minx->X()) minx = v;
	  if (v->X() > maxx->X()) maxx = v;
	  if (v->Y() < miny->Y()) miny = v;
	  if (v->Y() > maxy->Y()) maxy = v;
	  v = v->nextv; 
      }
}



long vertex_degree(Vertex_ptr vertex)
{
    long deg = 0;
    HalfEdge_ptr base = 0,he = 0;
    base = he = vertex->vedge;

    do
      {
	  deg++;
	  he = mate(he)->nxt;
      }
    while (he != base);
    return deg;
}



Face_ptr find_external_face( Pslg& pslg)
{
    Vertex_ptr minx,miny,maxx,maxy;
    HalfEdge_ptr he,maxy_he,left,right;
    
    find_min_max_x_y(pslg,minx,miny,maxx,maxy);  // Finds extreme points.
      
      he = FirstInComing(maxy);
    find_horiz(maxy,left,right);
    
    if (right != 0) maxy_he = right;
    else if (left != 0) maxy_he = mate(left)->nxt;
    else maxy_he = he->nxt;
    
    return maxy_he->wloop->lface; 
}


