/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * Ray2D.C - Ray2D & GraphicRay2D methods.
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    07.01.92  13:40 	Welson R. Jacometti  	Arquivo criado
 *
 */

#include "Ray2D.h"
#include "utilities.h"

GraphicObj* Ray2D::new_GraphicObj() {
    return new GraphicRay2D(this);
}

bool Ray2D::intersect( Ray2D &r)  {
    if (!Line2D::intersect(r))
      return false;

    Point2D res = Line2D::intersection(r);
    // calcula o valor de t no ponto de intersecao
    double t1, t2;
    if (horizontal())
      t1 = (res.X() - x0) / f;
    else
      t1 = (res.Y() - y0) / g;
    if (r.horizontal())
      t2 = (res.X() - r.x0) / r.f;
    else
      t2 = (res.Y() - r.y0) / r.g;
      
    if (t1 < 0 || t2 < 0)
      return false;
    
    return true;
}

bool Ray2D::intersect( Line2D &l)  { 
    if (!Line2D::intersect(l))
      return false;

    Point2D res = Line2D::intersection(l);
    double tr;
    if (horizontal())
      tr = (res.X() - x0) / f;
    else
      tr = (res.Y() - y0) / g;

    return tr >= 0;
}

bool Ray2D::intersect( Point2D &a,  Point2D &b)  {
    if (a.at_infinity() && !b.at_infinity())
      return intersect(Ray2D(b, a));

    if (!a.at_infinity() && b.at_infinity())
      return intersect(Ray2D(a, b));
    
    Ray2D l(a, b);    
    if (!Line2D::intersect(l))
      return false;

    Point2D res = Line2D::intersection(l);
    double tr;
    if (horizontal())
      tr = (res.X() - x0) / f;
    else
      tr = (res.Y() - y0) / g;

    if (a.at_infinity() && b.at_infinity())
      return tr >= 0;

    double tl;
    if (l.horizontal())
      tl = (res.X() - l.x0) / l.f;
    else
      tl = (res.Y() - l.y0) / l.g;

    return (tr >= 0 && tl >= 0 && tl <= 1);
}

Point2D Ray2D::intersection( Ray2D &r)  {
    Point2D res;
    
    if (intersect(r))
      res = Line2D::intersection(r);
    else
      res.set_at_infinity(0, 0);
      
    return res;
}

Point2D Ray2D::intersection( Line2D &l)  {
    Point2D res;
    
    if (intersect(l))
      res = Line2D::intersection(l);
    else
      res.set_at_infinity(0, 0);
     
    return res;
}

Point2D Ray2D::intersection( Point2D &a,  Point2D &b)  {
    Point2D res;
    
    if (intersect(a, b))
      res = Line2D::intersection(Line2D(a, b));
    else
      res.set_at_infinity(0, 0);
      
    return res;
}

IntersectType Ray2D::circle_intersections( Point2D &center, double radius, Point2D &i1, Point2D &i2)  {
    IntersectType line_result = Line2D::circle_intersections(center, radius, i1, i2);
    if (line_result == NoIntersection || line_result == Tangent)
      return line_result;

    // duas intersecoes - falta agora eliminar a(s) que nao interessa(m)
    double ti1, ti2;
    if (horizontal()) {
      ti1 = (i1.X() - x0) / f;
      ti2 = (i2.X() - x0) / f;
    }
    else {
      ti1 = (i1.Y() - y0) / g;
      ti2 = (i2.Y() - y0) / g;
    }

    if (ti1 < 0 && ti2 < 0)
      return NoIntersection;
      
    if (ti1 >= 0 && ti2 >= 0)
      return Intersect;
      
    if (ti1 < 0)
      i1 = i2;
      
    return JustOneIntersection;
}

void Ray2D::unfill_handles(World *w, bool show_handle)  {
    if (show_handle) {
      p1.unfill_handles(w);
      p2.unfill_handles(w);
    }
}

void Ray2D::draw_handles(World *w, GC gc, bool highlight, bool show_handle)  {
    if (show_handle) {
      p1.draw_handles(w, gc, highlight);
      p2.draw_handles(w, gc, highlight);
    }
}

void Ray2D::draw(World *w, GC gc)  {
    Point2D i1, i2;
    double x1, y1, x2, y2;
    IntersectType res;
        
    w->get_values(x1, y1, x2, y2);
    res = circle_intersections(Point2D((x1 + x2)/2, (y1 + y2)/2), 2 * fabs(((x1 + x2) / 2) - x1) , i1, i2);
    if (res == Intersect)
      w->DrawLine(gc, i1.X(), i1.Y(), i2.X(), i2.Y());
    else
      if (res == JustOneIntersection)
        w->DrawLine(gc, p1.X(), p1.Y(), i1.X(), i1.Y());
}

double Ray2D::distance_to_point(double x, double y)  {
    // calcula o valor de t no ponto mais proximo da linha
    double xj0 = x - x0;
    double yj0 = y - y0;
    double fgsq = (f * f) + (g * g);
    if (fgsq < EPSLON) {
      warning("Ray2D.C", "distance_to_point", "ray improperly defined");
      return TOO_FAR;
    }
    else {
      double tj = ((f * xj0) + (g * yj0)) * (1 / fgsq);
      if (tj >= 0)
        return Line2D::distance_to_point(x, y);
      else
        return p1.distance_to_point(x, y);
    }
}
