/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * World.C - implementacao dos metodos da classe World.
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    04.10.91  17:45 	Welson R. Jacometti  	Arquivo criado
 *
 */

#include "World.h"

/****************************************************************************

  Funcoes de interceptacao de call-backs e associacao `a instancia correta

*****************************************************************************/

static Attr_attribute WORLD_INSTANCE = xv_unique_key();

static void world_repaint_proc(Canvas, Xv_Window xv_win) {
    World *which_world = (World *)xv_get(xv_win, XV_KEY_DATA, WORLD_INSTANCE);

    which_world->call_repaint_proc();
}

static void world_event_proc(Xv_Window xv_win, Event *event) {
    World *which_world = (World *)xv_get(xv_win, XV_KEY_DATA, WORLD_INSTANCE);

    which_world->call_event_proc(event);
}

/****************************************************************************

                Implementacao dos metodos da classe World    

*****************************************************************************/

/*
 * Menor e maior diferencas permitidas entre valores minimos e maximos de 
 * um sistema de coordenadas.
 *
 */
 
#define	MIN_ACCURACY	0.5
#define MAX_ACCURACY	10E+5

/*
 * Valores que identificam QUEM inseriu um elemento na pilha de worlds.
 *
 */
 
#define ZOOM_IN		-1
#define INIT		0
#define ZOOM_OUT	1

World::World(Canvas cv) {
    init(cv, 0, (double)xv_get(canvas_paint_window(cv), XV_HEIGHT), 
                (double)xv_get(canvas_paint_window(cv), XV_WIDTH), 0);
}

World::World(Canvas cv, double x1, double y1, double x2, double y2) {
    init(cv, x1, y1, x2, y2);
}

World::~World() {
    if (pixmap)
      FreePixmap(pixmap);	// libera pixmap caso ainda exista
      
    while(remove_world());
    while(remove_event_proc());
}

void World::restart() {
    reset();
    (void)define();
    initial = *world;    
}

bool World::define() {
    return define(0, (double)xv_get(canvas_paint_window(canvas), XV_HEIGHT),
                     (double)xv_get(canvas_paint_window(canvas), XV_WIDTH), 0);
}
    
bool World::define(double x1, double y1, double x2, double y2) {
    double difx = fabs(x1 - x2);
    double dify = fabs(y1 - y2);
    
    if (difx <= MIN_ACCURACY || difx >= MAX_ACCURACY || dify <= MIN_ACCURACY || dify >= MAX_ACCURACY)
      return false; // range too small/big for a new coordinate system
    
    world->min_x = x1;
    world->min_y = y1;
    world->max_x = x2;
    world->max_y = y2;
    
    world->view_width  = (int)xv_get(canvas, CANVAS_WIDTH);
    world->view_height = (int)xv_get(canvas, CANVAS_HEIGHT);
    
    world->paint_width  = (int)xv_get(xv_win, XV_WIDTH);
    world->paint_height = (int)xv_get(xv_win, XV_HEIGHT);
    
    world->theta_x = world->paint_width  / (world->max_x - world->min_x);
    world->theta_y = world->paint_height / (world->max_y - world->min_y);

    return true;
}
      
void World::get_values(double &x1, double &y1, double &x2, double &y2) {
    x1 = world->min_x;
    y1 = world->min_y;
    x2 = world->max_x;
    y2 = world->max_y;
}

void World::get_view(int &width, int &height) {
    width = world->view_width;
    height = world->view_height;
}

void World::get_paint(int &width, int &height) {
    width = world->paint_width;
    height = world->paint_height;
}

void World::compress_drag(Event *event, int to_retain) {
    XEvent current = *(event->ie_xevent);
    XEvent ahead;
    
    while (XEventsQueued(current.xmotion.display, QueuedAfterReading) > to_retain) { 
      XPeekEvent(current.xmotion.display, &ahead);
      if (ahead.type != MotionNotify || ahead.xmotion.window != current.xmotion.window)
        break;
      window_read_event(xv_win, event);
      current = *(event->ie_xevent);
    }
}
    
double World::string_width(char *str) {
    Font_string_dims dims;
    (void)xv_get((Xv_Font)xv_get(canvas, XV_FONT), FONT_STRING_DIMS, str, &dims);
    return dims.width;
}

void World::DrawCenterString(GC gc, double x, double y, char *str, int strlen) {
    Font_string_dims dims;
    int w_div_2;
    (void)xv_get((Xv_Font)xv_get(canvas, XV_FONT), FONT_STRING_DIMS, str, &dims);
    w_div_2 = dims.width / 2;
    XDrawString(dpy, xwin, gc, xcoord(x) - w_div_2, ycoord(y), str, strlen);
}

void World::reset() {
    while (world->next)
      (void)remove_world();
     
    *world = initial;
}

bool World::zoom_in(double x, double y, double factor) {
    double dx, dy;
        
    if (factor < 1) {
      cerr << "world.C: zoom_in: invalid zoom factor: " << factor << endl;
      return false;
    }

    dx = (world->max_x - world->min_x) / (2 * factor);
    dy = (world->max_y - world->min_y) / (2 * factor);
    
    return define(x - dx, y - dy, x + dx, y + dy);
}

bool World::zoom_out(double x, double y, double factor) {
    double dx, dy;
        
    if (factor < 1) {
      cerr << "world.C: zoom_out: invalid zoom factor: " << factor << endl;
      return false;
    }
    
    dx = ((world->max_x - world->min_x) / 2) * factor;
    dy = ((world->max_y - world->min_y) / 2) * factor;

    return define(x - dx, y - dy, x + dx, y + dy);
}

bool World::zoom_in_stack(double x, double y, double factor) {
    if (world->inserted_by == ZOOM_OUT)
      zoom_previous();
    else {
      WorldData *new_world = new WorldData;
      assert(new_world);
        
      if (!factor) {
        cerr << "world.C: zoom_in_stack: invalid zoom factor" << endl;
        return false;
      }

      *new_world = *world;
      insert_world(new_world);
      if (!zoom_in(x, y, factor)) {
        (void)remove_world();
        return false;
      }
      else {
        world->inserted_by = ZOOM_IN;
        return true;
      }
    }
}

bool World::zoom_out_stack(double x, double y, double factor) {
    if (world->inserted_by == ZOOM_IN) 
      zoom_previous();
    else {
      WorldData *new_world = new WorldData;
      assert(new_world);
        
      if (!factor) {
        cerr << "world.C: zoom_out_stack: invalid zoom factor" << endl;
        return false;
      }

      *new_world = *world;
      insert_world(new_world);
      if (!zoom_out(x, y, factor)) {
        (void)remove_world();
        return false;
      }
      else {
        world->inserted_by = ZOOM_OUT;
        return true;
      }
    }
}

bool World::zoom_previous() {
    if (world->next) {
      (void)remove_world();
      return true;
    }
    else
      return false;
}

void World::pscroll_begin(GC gc) {
    if (!pixmap) {
      pixmap = CreatePixmap(xwin, world->paint_width, world->paint_height, DefaultDepth(dpy, 0));
      CopyArea(xwin, pixmap, gc, 0, 0, world->paint_width, world->paint_height, 0, 0);
    }
}

void World::pscroll_end(int final_x, int final_y) {
    if (pixmap) {
      FreePixmap(pixmap);
      pixmap = 0;
      scroll(final_x, final_y);
    }
}

void World::pscroll_same_size(GC gc, int to_x, int to_y) {
    static int src_x, src_y, width, height;
    
    if (pixmap) {
      if (to_x >= 0 && to_y >= 0) {
        XFillRectangle(dpy, xwin, gc, 0, 0, to_x, world->view_height);
        XFillRectangle(dpy, xwin, gc, 0, 0, world->view_width, to_y);
        width = world->view_width - to_x;
        height = world->view_height - to_y;        
        src_x = src_y = 0;
      }
      else
        if (to_x >= 0 && to_y < 0) {
          XFillRectangle(dpy, xwin, gc, 0, 0, to_x, world->view_height);
          XFillRectangle(dpy, xwin, gc, 0, world->view_height + to_y, world->view_width, -to_y);
          width = world->view_width - to_x;
          height = world->view_height + to_y;
          src_x = 0; 
          src_y = -to_y; to_y = 0;
        }
      else
        if (to_x < 0 && to_y >= 0) {
          XFillRectangle(dpy, xwin, gc, 0, 0, world->view_width, to_y);
          XFillRectangle(dpy, xwin, gc, world->view_width + to_x, 0, -to_x, world->view_height);
          width = world->view_width + to_x;
          height = world->view_height - to_y;
          src_x = -to_x; to_x = 0;
          src_y = 0;
        }
      else /* to_x < 0 && to_y < 0 */ {
          XFillRectangle(dpy, xwin, gc, 0, world->view_height + to_y, world->view_width, -to_y);
          XFillRectangle(dpy, xwin, gc, world->view_width + to_x, 0, -to_x, world->view_height);
          width = world->view_width + to_x;
          height = world->view_height + to_y;
          src_x = -to_x; to_x = 0;
          src_y = -to_y; to_y = 0;
      } 

      CopyArea(pixmap, xwin, gc, src_x, src_y, width, height, to_x, to_y);
    }	       
}

void World::pscroll_paint_bigger(GC gc, int to_x, int to_y) {
    static int src_x, src_y, width, height;
    
    width = world->view_width;
    height = world->view_height;
    
    if (pixmap) {
      if (to_x >= 0 && to_y >= 0) {
        width = width - to_x;
        height = height - to_y;        
        src_x = src_y = 0;
      }
      else
        if (to_x >= 0 && to_y < 0) {
          width = width - to_x;
          src_x = 0; 
          src_y = -to_y; to_y = 0;
        }
      else
        if (to_x < 0 && to_y >= 0) {
          height = height - to_y;
          src_x = -to_x; to_x = 0;
          src_y = 0;
        }
      else /* to_x < 0 && to_y < 0 */ {
          src_x = -to_x; to_x = 0;
          src_y = -to_y; to_y = 0;
      } 

      XFillRectangle(dpy, xwin, gc, 0, 0, world->view_width, world->view_height);
      CopyArea(pixmap, xwin, gc, src_x, src_y, width, height, to_x, to_y);
    }	       
}

void World::scroll(int horz_amount, int vert_amount) {
    double horz, vert;

    horz = horz_amount * ratio_x();
    vert = vert_amount * ratio_y();
    define(world->min_x - horz, world->min_y - vert,
    	   world->max_x - horz, world->max_y - vert);
}

void World::set_repaint_proc(f_repaint r_proc) {
    repaint_proc = r_proc;
    (void)xv_set(canvas, CANVAS_REPAINT_PROC, world_repaint_proc, NULL);
}

void World::set_event_proc(f_events e_proc) {
    EventFunctions *new_node;
    
    while (remove_event_proc());
    
    new_node = new EventFunctions;
    assert(new_node);
    new_node->event_proc = e_proc;
    insert_event_proc(new_node);
    (void)xv_set(xv_win, WIN_EVENT_PROC, world_event_proc, NULL);
}

void World::reset_event_proc() {
    while (event_functions && event_functions->next)
      (void)remove_event_proc();
}

void World::add_event_proc(f_events e_proc) {
    EventFunctions *new_node;
    
    new_node = new EventFunctions;
    assert(new_node);
    new_node->event_proc = e_proc;
    insert_event_proc(new_node);
}

void World::del_event_proc() {
    if (event_functions && event_functions->next)
      (void)remove_event_proc();    
}

void World::call_event_proc(Event *event) {
    EventFunctions *level = event_functions;

    while (level) {
      if (level->event_proc(this, xv_win, event))
        break;
      level = level->next;
    }
}
    
// Private function members

void World::init(Canvas cv, double x1, double y1, double x2, double y2) {
    canvas = cv;	
    xv_win = canvas_paint_window(canvas);
    dpy = (Display *)xv_get(canvas, XV_DISPLAY);
    xwin = (Window)xv_get(xv_win, XV_XID);

    (void)xv_set(xv_win, XV_KEY_DATA, WORLD_INSTANCE, this, NULL);
    
    pixmap = 0;
    repaint_proc = 0;
    event_functions = 0;

    world = new WorldData;
    assert(world);
    world->inserted_by = INIT;
    if (!define(x1, y1, x2, y2)) {
      cerr << "world.C: init: bad initial values for a coordinate system" << endl;
      cerr << "               x1 = " << x1 << endl;
      cerr << "               y1 = " << y1 << endl;
      cerr << "               x2 = " << x2 << endl;
      cerr << "               y2 = " << y2 << endl;
      exit(1);
    }
    initial = *world;
}

void World::insert_world(WorldData *data) {
    data->next = world;
    world = data;    
}

World::WorldData* World::remove_world() {
    if (world) {	
      WorldData *save = world;
      world = world->next;
      delete save;
    }
    return world;
}

void World::insert_event_proc(EventFunctions *new_node) {
    new_node->next = event_functions;
    event_functions = new_node;
}

World::EventFunctions *World::remove_event_proc() {
    if (event_functions) {
      EventFunctions *save = event_functions;
      event_functions = event_functions->next;
      delete save;
    }
    return event_functions;
}


