/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * World.h - contem a definicao da classe World, que define um sistema de
 *           coordenadas virtual associada a um canvas.
 *
 *           link flag (-lm)
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    04.10.91  17:45 	Welson R. Jacometti  	Arquivo criado
 *    22.10.91  11:30	Welson			Funcao de compressao de eventos
 *    08.11.91  20:05	Welson			zoom_in/out_stack
 *    13.11.91  15:00	Welson			repaint/events
 *    11.02.92  10:30   Welson			dynamic external event_handlers
 *
 */

#ifndef _WORLD_DEFINED
#define _WORLD_DEFINED

#include <iostream.h>
#include <math.h>
#include <X11/X.h>
#include <X11/Xlib.h>   
#include <X11/Xutil.h> 
#include <xview/xview.h>
#include <xview/canvas.h>
#include <xview/font.h>
#include <assert.h>
#include "global.h" 

#define MAX_WH	1700	// max width/height supported by X

class World {
public:
    // constructors
    World(Canvas);
    World(Canvas, double x1, double y1, double x2, double y2);

    // destructor
    ~World();

    // to avoid the use of World(const World&) like initialization and
    // assignments between `Worlds', the following two methods issue
    // error messages and cause abnormal program termination
    World(const World&) {
      cerr << "Fatal error: World.h: World(const World&) initialization is prohibited for the World class" << endl;
      exit(1);
    }
    
    World& operator = (const World&) {
      cerr << "Fatal error: World.h: Assignment between `Worlds' is prohibited for the World class" << endl;
      exit(1);
      return *this; 	// for no compiler warnings
    }
    
    // (re)definition functions
    void restart();
    bool define();
    bool define(double x1, double y1, double x2, double y2);
    
    // query functions
    void get_values(double &x1, double &y1, double &x2, double &y2);
    void get_view(int &width, int &height);
    void get_paint(int &width, int &height);
    Canvas get_canvas() { return canvas; }
    Xv_Window get_paint() { return xv_win; }
    Display *get_display() { return dpy; }
    Window get_window() { return xwin; }
    
    // translation functions
    int xcoord(double virtual_x) {
      return rint(world->theta_x *(virtual_x - world->min_x));
    }

    int ycoord(double virtual_y) {
      return rint(world->theta_y *(virtual_y - world->min_y));
    }

    double virtual_xcoord(int real_x) {
      return (double)((real_x / world->theta_x) + world->min_x);
    }

    double virtual_ycoord(int real_y) {
      return (double)((real_y / world->theta_y) + world->min_y);
    }

    // draw functions to substitute XLib draw functions
    void Flush() {
      XFlush(dpy);
    }    

    void Sync() {
      XSync(dpy, 0);
    }
    
    Pixmap CreatePixmap(Drawable drawable, int width, int height, int depth) {
      return XCreatePixmap(dpy, drawable, width, height, depth);
    }
    
    void FreePixmap(Pixmap pixmap) {
      XFreePixmap(dpy, pixmap);
    }
    
    void CopyArea(Drawable src, Drawable dest, GC gc, int src_x, int src_y, int width, int height, int dest_x, int dest_y) {
      XCopyArea(dpy, src, dest, gc, src_x, src_y, width, height, dest_x, dest_y);
    }
    
    void ClearWindow() {
      XClearWindow(dpy, xwin);
    }
	
    void DrawPoint(GC gc, double x, double y) {
      XDrawPoint(dpy, xwin, gc, xcoord(x), ycoord(y));
    }

    void DrawLine(GC gc, double x1, double y1, double x2, double y2) {
      XDrawLine(dpy, xwin, gc, xcoord(x1), ycoord(y1), xcoord(x2), ycoord(y2));
    }

    void DrawRectangle(GC gc, double x1, double y1, double x2, double y2) {
      regularize(x1, y1, x2, y2);
      unsigned int w = (unsigned int)(x2 - x1);
      unsigned int h = (unsigned int)(y2 - y1);
      if (w > MAX_WH || h > MAX_WH)
        cerr << "Warning: World.h: rectangle too big, can't draw" << endl;
      else
        XDrawRectangle(dpy, xwin, gc, (int)x1, (int)y1, (short)w, (short)h);
    }

    void FillRectangle(GC gc, double x1, double y1, double x2, double y2) {
      regularize(x1, y1, x2, y2);
      unsigned int w = (unsigned int)(x2 - x1);
      unsigned int h = (unsigned int)(y2 - y1);
      if (w > MAX_WH || h > MAX_WH)
        cerr << "Warning: World.h: rectangle too big, can't draw" << endl;
      else
        XFillRectangle(dpy, xwin, gc, (int)x1, (int)y1, (short)w, (short)h);
    }

    void DrawArc(GC gc, double x1, double y1, double x2, double y2, int arc_i = 0, int arc_f = 23040 /* 360 x 64 */) {
      regularize(x1, y1, x2, y2);
      unsigned int w = (unsigned int)(x2 - x1);
      unsigned int h = (unsigned int)(y2 - y1);
      if (w > MAX_WH || h > MAX_WH)
        cerr << "Warning: World.h: arc too big, can't draw" << endl;
      else
        XDrawArc(dpy, xwin, gc, (int)x1, (int)y1, (short)w, (short)h, arc_i, arc_f);
    }

    void DrawArc(GC gc, double x, double y, double radius, int arc_i = 0, int arc_f = 23040) {
      DrawArc(gc, x - radius, y - radius, x + radius, y + radius, arc_i, arc_f);
    }

    void FillArc(GC gc, double x1, double y1, double x2, double y2, int arc_i = 0, int arc_f = 23040) {
      regularize(x1, y1, x2, y2);
      unsigned int w = (unsigned int)(x2 - x1);
      unsigned int h = (unsigned int)(y2 - y1);
      if (w > MAX_WH || h > MAX_WH)
        cerr << "Warning: World.h: arc too big, can't draw" << endl;
      else
        XFillArc(dpy, xwin, gc, (int)x1, (int)y1, (short)w, (short)h, arc_i, arc_f);
    }

    void FillArc(GC gc, double x, double y, double radius, int arc_i = 0, int arc_f = 23040) {
      FillArc(gc, x - radius, y - radius, x + radius, y + radius, arc_i, arc_f);
    }

    void DrawString(GC gc, double x, double y, char *str, int strlen) {
      XDrawString(dpy, xwin, gc, xcoord(x), ycoord(y), str, strlen);
    }

    double string_width(char *str);

    void DrawCenterString(GC gc, double x, double y, char *str, int strlen);

    XFontStruct *LoadQueryFont(char *name) {
      return XLoadQueryFont (dpy, name);
    }

    void SetFont(GC gc, Font font) {
      XSetFont (dpy, gc, font);
    }

    // GC related functions
    GC defaultGC() {
      return DefaultGC(dpy, DefaultScreen(dpy));
    }

    GC CreateGC(unsigned long mask, XGCValues &values) {
      return XCreateGC(dpy, xwin, mask, &values);
    }

    GC CreateGC(unsigned long mask, XGCValues *values) {
      return XCreateGC(dpy, xwin, mask, values);
    }
    
    void ChangeGC(GC gc, unsigned long mask, XGCValues &values) {
      XChangeGC(dpy, gc, mask, &values);
    }

    void CopyGC(GC gcsource, unsigned long mask, GC gcdest) {
      XCopyGC(dpy, gcsource, mask, gcdest);
    }

    void FreeGC(GC gc) {
      XFreeGC(dpy, gc);
    }

    void SetForeground(GC gc, unsigned long color) {
      XSetForeground(dpy, gc, color);
    }
    
    void SetBackground(GC gc, unsigned long color) {
      XSetBackground(dpy, gc, color);
    }
    
    void SetFunction(GC gc, int function) {
      XSetFunction(dpy, gc, function);
    }
    
    void SetPlaneMask(GC gc, unsigned long mask) {
      XSetPlaneMask(dpy, gc, mask);
    }

    void SetLineAttributes(GC gc, int width, int style, int cap, int join) {
      XSetLineAttributes(dpy, gc, width, style, cap, join);
    }

    // zoom reset function
    void reset();
    
    // zoom functions
    bool zoom_in(double x, double y, double factor);
    bool zoom_out(double x, double y, double factor);	
    bool zoom_in_stack(double x, double y, double factor);
    bool zoom_out_stack(double x, double y, double factor);
    bool zoom_previous();

    // auxiliary zoom functions
    double ratio_x() { return (world->max_x - world->min_x) / fabs(initial.max_x - initial.min_x); }
    double ratio_y() { return (world->max_y - world->min_y) / fabs(initial.max_y - initial.min_y); }
    
    // scroll functions that manipulate pixmaps
    void pscroll_begin(GC);	
    void pscroll_end(int final_x, int final_y);
    void pscroll_same_size(GC gc, int to_x, int to_y);	  // to be used when paint window size == view window size
    void pscroll_paint_bigger(GC gc, int to_x, int to_y); // to be used when paint window is bigger than the view window (usually lens)
    
    // scroll function that adjusts only world settings
    void scroll(int horz_amount, int vert_amount);

    // repaint & events_proc facilities
    typedef void (*f_repaint)(World *);
    typedef bool (*f_events)(World *, Xv_Window, Event *);

    // function to compress LOC_DRAG events associated with the world canvas
    void compress_drag(Event *event, int to_retain = 0);
    
    void set_repaint_proc(f_repaint);
    f_repaint get_repaint_proc() { return repaint_proc; }
    void call_repaint_proc() { repaint_proc(this); }

    void set_event_proc(f_events);
    void reset_event_proc();
    void add_event_proc(f_events);
    void del_event_proc();
    void call_event_proc(Event *);

    class EventFunctions {
    public:
      EventFunctions() { next = 0; }

      f_events event_proc;
      EventFunctions *next;
    };
					    
    class WorldData {
    public:
        WorldData() { next = 0; }

        int inserted_by; 
        double min_x, min_y, max_x, max_y, theta_x, theta_y;
        int paint_width, paint_height;
        int view_width, view_height;
        WorldData *next;
    };

private:
    
    // data initialization & structure functions
    void init(Canvas cv, double x1, double y1, double x2, double y2);    
    void insert_world(WorldData *);
    WorldData* remove_world();
         
    // event functions related routines
    void insert_event_proc(EventFunctions *);
    EventFunctions *remove_event_proc();
	        
    // auxiliary draw functions
    void swap(double &x1, double &x2) {
      double tmp = x1;
      x1 = x2; 
      x2 = tmp;
    }

    void regularize(double &x1, double &y1, double &x2, double &y2) {
      x1 = xcoord(x1);
      y1 = ycoord(y1);
      x2 = xcoord(x2);
      y2 = ycoord(y2);
      if (x1 > x2) swap(x1, x2);
      if (y1 > y2) swap(y1, y2);
    }

    // internal data representation
    Canvas canvas;
    Xv_Window xv_win;
    Display *dpy;
    Window xwin; 
    Pixmap pixmap;
    WorldData initial;
    WorldData *world; 

    f_repaint repaint_proc;
    EventFunctions *event_functions;
};

#endif ~_WORLD_DEFINED;




