/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 * objects.C - manutencao dos objetos geometricos manipulados pelo GeoLab.
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    12.01.92  18:25 	Welson R. Jacometti  	Arquivo criado
 *
 */

#include "objects.h"
#include "algorithms.h"		// este tipo de dependencia circular e' perfeitamente possivel
#include "color.h"		// e neste caso necessaria para manter a separacao logica dos modulos 
#include "line_settings.h"
#include "edit_mode.h"

bool HandlesOn = true;

#define MIN_DIST_PIXELS	3	// minimal distance for select by distance

list(GraphicObj_ptr) GRAPHIC_OBJECTS;
list(GraphicObj_ptr) UNDELETABLE_OBJECTS;
list(list_item) SELECTED_OBJECTS;

void objects_insert(GraphicObj *obj) {
    GRAPHIC_OBJECTS.push(obj);
}

void objects_insert_and_select(GraphicObj *obj) {
    obj->select();
    SELECTED_OBJECTS.init_iterator();
    SELECTED_OBJECTS.push(GRAPHIC_OBJECTS.push(obj));
}

void objects_insert_external(GraphicObj *obj, char *function, char *module) {
    obj->set_external(function, module);
    GRAPHIC_OBJECTS.push(obj);
}

void objects_clear() {
    GraphicObj *obj;

    forall(obj, GRAPHIC_OBJECTS) {
      delete obj;
    }

    GRAPHIC_OBJECTS.clear();
    SELECTED_OBJECTS.init_iterator();
    SELECTED_OBJECTS.clear();
}

bool objects_empty() {
    return GRAPHIC_OBJECTS.empty();
}

bool objects_selected_empty() {
    return SELECTED_OBJECTS.empty();
}

void objects_repaint(World *w) {
    GraphicObj *obj;

    if (!GRAPHIC_OBJECTS.empty()) {
      GRAPHIC_OBJECTS.set_iterator(GRAPHIC_OBJECTS.last());
      while (GRAPHIC_OBJECTS.current_element(obj)) {
      	obj->draw(w, CopyMode);
      	GRAPHIC_OBJECTS.move_iterator(backward);
      }
    }
}

void objects_repaint_all_but(GraphicObj *not_to_repaint, World *w) {
    GraphicObj *obj;

    if (!GRAPHIC_OBJECTS.empty()) {
      GRAPHIC_OBJECTS.set_iterator(GRAPHIC_OBJECTS.last());
      while (GRAPHIC_OBJECTS.current_element(obj)) {
      	if (obj != not_to_repaint)
          obj->draw(w, CopyMode);
      	GRAPHIC_OBJECTS.move_iterator(backward);
      }
    }
}

void objects_repaint_unselected(World *w) {
    GraphicObj *obj;

    if (!GRAPHIC_OBJECTS.empty()) {
      GRAPHIC_OBJECTS.set_iterator(GRAPHIC_OBJECTS.last());
      while (GRAPHIC_OBJECTS.current_element(obj)) {
      	if (!obj->selected())
          obj->draw(w, CopyMode);
      	GRAPHIC_OBJECTS.move_iterator(backward);
      }
    }
}

void objects_refresh(World *w) {
    Frame frame = (Frame)xv_get(w->get_canvas(), XV_OWNER);
    (void)xv_set(frame, FRAME_LEFT_FOOTER, "", 0);
    (void)xv_set(frame, FRAME_RIGHT_FOOTER, "", 0);
    w->ClearWindow();
    objects_repaint(w);
}

void objects_set_handle(World *w, bool all, bool show) {
    bool did_something = false;
    
    if (all) {
      GraphicObj *obj;
      forall(obj, GRAPHIC_OBJECTS) {
      	obj->set_handle(show);
      	did_something = true;
      }
    }
    else {
      list_item it;
      forall(it, SELECTED_OBJECTS) {
        GRAPHIC_OBJECTS[it]->set_handle(show);
        did_something = true;
      }
    }
    
    if (did_something)    
      objects_refresh(w);
}

bool objects_select_one_left_button(World *w, double x, double y) {
    list_item it;
    GraphicObj *obj;
    bool gotcha = false;
        
    forall_list_items(it, GRAPHIC_OBJECTS) {
      obj = GRAPHIC_OBJECTS.contents(it);
      if (obj->select_test(x, y)) {
        gotcha = true;
        break;
      }
    }

    if (!gotcha) 	// now let's try by distance (handles have priority)
      forall_list_items(it, GRAPHIC_OBJECTS) {
        obj = GRAPHIC_OBJECTS.contents(it);
        if (obj->distance_to_point(x, y) <= MIN_DIST_PIXELS) {
          gotcha = true;
          break;
        }
      }

    if (gotcha) {
      if (!obj->selected()) {
      	objects_unselect(w);
       	obj->select();
        SELECTED_OBJECTS.push(it);
        objects_repaint(w);
      }
      return true;
    }
    else {
      if (!SELECTED_OBJECTS.empty()) {
        objects_unselect(w);
        objects_repaint(w);
      }  
    }
    
    return false;	// everybody unselected
}

bool objects_select_one_middle_button(World *w, double x, double y) {
    list_item it;
    GraphicObj *obj;
        
    forall_list_items(it, SELECTED_OBJECTS) {
      obj = GRAPHIC_OBJECTS.contents(SELECTED_OBJECTS.contents(it));
      if (obj->select_test(x, y)) {
      	obj->unselect();
      	obj->unfill_handles(w);
        SELECTED_OBJECTS.init_iterator();
        SELECTED_OBJECTS.del_item(it);
        objects_repaint(w);
        return true;	// nothing more to do
      }
    }
    
    forall_list_items(it, GRAPHIC_OBJECTS) {
      obj = GRAPHIC_OBJECTS.contents(it);
      if (obj->select_test(x, y)) {
      	obj->select();
	SELECTED_OBJECTS.push(it);
      	objects_repaint(w);
        GRAPHIC_OBJECTS.init_iterator();
        return true;	// nothing more to do
      }
    }

    // now let's try by distance (handles have priority)

    forall_list_items(it, SELECTED_OBJECTS) {
      obj = GRAPHIC_OBJECTS.contents(SELECTED_OBJECTS.contents(it));
      if (obj->distance_to_point(x, y) <= MIN_DIST_PIXELS) {
      	obj->unselect();
      	obj->unfill_handles(w);
        SELECTED_OBJECTS.init_iterator();
        SELECTED_OBJECTS.del_item(it);
        objects_repaint(w);
        return true;	// nothing more to do
      }
    }
    
    forall_list_items(it, GRAPHIC_OBJECTS) {
      obj = GRAPHIC_OBJECTS.contents(it);
      if (obj->distance_to_point(x, y) <= MIN_DIST_PIXELS) {
      	obj->select();
	SELECTED_OBJECTS.push(it);
      	objects_repaint(w);
        GRAPHIC_OBJECTS.init_iterator();
        return true;	// nothing more to do
      }
    }

    return false;	// did nothing
}

void objects_select_all(World *w) {
    list_item it;
    GraphicObj *obj;
    bool selected_some = false;
      
    SELECTED_OBJECTS.init_iterator();
    
    forall_list_items(it, GRAPHIC_OBJECTS) {
      obj = GRAPHIC_OBJECTS.contents(it);
      if (!obj->selected()) {
      	obj->select();
        SELECTED_OBJECTS.push(it);
        selected_some = true;
      }
    }
    
    if (selected_some)
      objects_repaint(w);
}

void objects_select(World *w, double x1, double y1, double x2, double y2, bool FLIP) {
    list_item it;
    GraphicObj *obj;
    list(list_item) ToBeRemoved;
    bool did_something = false;

    if (FLIP) {
      forall_list_items(it, SELECTED_OBJECTS) {
        obj = GRAPHIC_OBJECTS.contents(SELECTED_OBJECTS.contents(it));
        if (obj->select_test(x1, y1, x2, y2))
          ToBeRemoved.push(it);
          did_something = true;
      }
    }

    SELECTED_OBJECTS.init_iterator();
    forall_list_items(it, GRAPHIC_OBJECTS) {
      obj = GRAPHIC_OBJECTS.contents(it);
      if (!obj->selected() && obj->select_test(x1, y1, x2, y2)) {
      	obj->select();
        SELECTED_OBJECTS.push(it);
        did_something = true;
      }
    }
    
    if (FLIP) {
      forall(it, ToBeRemoved) {
        obj = GRAPHIC_OBJECTS.contents(SELECTED_OBJECTS.contents(it));
        SELECTED_OBJECTS.del_item(it);
        obj->unselect();
        obj->unfill_handles(w);
        did_something = true;
      }
      ToBeRemoved.clear();
    }
     
    if (did_something)   
      objects_repaint(w);
}
    
void objects_unselect(World *w) {
    list_item it;

    forall(it, SELECTED_OBJECTS) {
      GRAPHIC_OBJECTS.contents(it)->unselect();
      GRAPHIC_OBJECTS.contents(it)->unfill_handles(w);
    }      
    SELECTED_OBJECTS.clear();
}
	    
void objects_unselect_no_refresh() {
    list_item it;

    forall(it, SELECTED_OBJECTS) {
      GRAPHIC_OBJECTS.contents(it)->unselect();
    }      
    SELECTED_OBJECTS.clear();
}
	    
void objects_set_color_selected(World *w, unsigned long fg, unsigned long bg) {
    list_item it;
    
    forall(it, SELECTED_OBJECTS) {
      GRAPHIC_OBJECTS.contents(it)->set_color(w, fg, bg);
    }      

    if (!SELECTED_OBJECTS.empty())
      objects_repaint(w);
}
	    
void objects_set_line_attr_selected(World *w, int lw, int ls) {
    list_item it;
    
    forall(it, SELECTED_OBJECTS) {
      GRAPHIC_OBJECTS.contents(it)->set_line_attr(w, lw, ls);
    }      

    if (!SELECTED_OBJECTS.empty())
      objects_refresh(w);
}
	    
void objects_delete_selected(World *w) {
    list_item it;
    GraphicObj *obj;
    list(list_item) aux;

    SELECTED_OBJECTS.init_iterator();
    if (!SELECTED_OBJECTS.empty()) {
      bool deleted_some = false;
      
      forall(obj, UNDELETABLE_OBJECTS) {
        delete obj;
      }
      UNDELETABLE_OBJECTS.clear();

      forall_list_items(it, SELECTED_OBJECTS) {
      	if (!GRAPHIC_OBJECTS[SELECTED_OBJECTS[it]]->anchored())
      	  aux.push(it);
      }      

      GRAPHIC_OBJECTS.init_iterator();
      forall(it, aux) {
        obj = GRAPHIC_OBJECTS.del_item(SELECTED_OBJECTS[it]);
        SELECTED_OBJECTS.del_item(it);
        UNDELETABLE_OBJECTS.push(obj);
        deleted_some = true;
      }
      
      if (deleted_some)
        objects_refresh(w);
    }
}
    
void objects_delete_selected_no_refresh() {
    list_item it;
    GraphicObj *obj;
    list(list_item) aux;

    SELECTED_OBJECTS.init_iterator();
    if (!SELECTED_OBJECTS.empty()) {
      
      forall(obj, UNDELETABLE_OBJECTS) {
        delete obj;
      }
      UNDELETABLE_OBJECTS.clear();

      forall_list_items(it, SELECTED_OBJECTS) {
      	if (!GRAPHIC_OBJECTS[SELECTED_OBJECTS[it]]->anchored())
      	  aux.push(it);
      }      

      GRAPHIC_OBJECTS.init_iterator();
      forall(it, aux) {
        obj = GRAPHIC_OBJECTS.del_item(SELECTED_OBJECTS[it]);
        SELECTED_OBJECTS.del_item(it);
        UNDELETABLE_OBJECTS.push(obj);
      }
    }
}
    
void objects_undelete_selected(World *w) {
    GraphicObj *obj;

    SELECTED_OBJECTS.init_iterator();
    if (!UNDELETABLE_OBJECTS.empty()) {
      forall(obj, UNDELETABLE_OBJECTS) {
        SELECTED_OBJECTS.push(GRAPHIC_OBJECTS.push(obj));
      }
      UNDELETABLE_OBJECTS.clear();
      objects_repaint(w);
    }
}
    
void objects_highlight(World *w) {
    list_item it;
    
    forall(it, SELECTED_OBJECTS) {
      GRAPHIC_OBJECTS.contents(it)->draw(w, CopyMode);
    }      
}

void objects_glue() {
    list_item it;
    
    forall(it, SELECTED_OBJECTS) {
      GRAPHIC_OBJECTS.contents(it)->glue();
    }      
}

void objects_unglue() {
    list_item it;
    
    forall(it, SELECTED_OBJECTS) {
      GRAPHIC_OBJECTS.contents(it)->unglue();
    }      
}

void objects_anchor() {
    list_item it;
    
    forall(it, SELECTED_OBJECTS) {
      GRAPHIC_OBJECTS.contents(it)->anchor();
    }      
}

void objects_unanchor() {
    list_item it;
    
    forall(it, SELECTED_OBJECTS) {
      GRAPHIC_OBJECTS.contents(it)->unanchor();
    }      
}

void objects_set_string(World *w, char *string) {
    list_item it;
    
    forall(it, SELECTED_OBJECTS) {
      GRAPHIC_OBJECTS.contents(it)->set_string(string);
    }
    
    objects_refresh(w);
}

void objects_dump() {
    list_item it;
    
    forall(it, SELECTED_OBJECTS) {
      GRAPHIC_OBJECTS.contents(it)->print();
    }      
}

GraphicObj* objects_object_selected_at(double x, double y, bool no_if_glued) {
    GraphicObj *obj;
    
    forall(obj, GRAPHIC_OBJECTS) {
      if (obj->select_test(x, y)) {
        GRAPHIC_OBJECTS.init_iterator();
        if (no_if_glued && obj->glued())
          return 0;
        else
          return obj;
      }
    }      

    // now let's try by distance (handles have priority)

    forall(obj, GRAPHIC_OBJECTS) {
      if (obj->distance_to_point(x, y) <= MIN_DIST_PIXELS) {
        GRAPHIC_OBJECTS.init_iterator();
        if (no_if_glued && obj->glued())
          return 0;
        else
          return obj;
      }
    }

    return 0;
}

GraphicObj* objects_object_selected_by_handle_at(double x, double y, bool no_if_glued) {
    GraphicObj *obj;
    
    forall(obj, GRAPHIC_OBJECTS) {
      if (obj->select_test(x, y)) {
        GRAPHIC_OBJECTS.init_iterator();
        if (no_if_glued && obj->glued())
          return 0;
        else
          return obj;
      }
    }      

    return 0;
}

bool objects_something_selected_at(double x, double y, bool no_if_glued) {
    list_item it;
    
    forall(it, SELECTED_OBJECTS) {
      if (GRAPHIC_OBJECTS[it]->select_test(x, y)) {
        SELECTED_OBJECTS.init_iterator();
        if (no_if_glued && GRAPHIC_OBJECTS[it]->glued())
          return false;
        else
          return true;
      }
    }      
    return false;    
}

bool objects_duplicate_selected(bool no_if_glued) {
    list_item it;			// retorna true se conseguiu duplicar
    GraphicObj *obj;			// com sucesso todos os objetos selecionados
    bool res = true;
    
    forall(it, SELECTED_OBJECTS) {
      if (!no_if_glued || (no_if_glued && !GRAPHIC_OBJECTS[it]->glued())) {
        obj = GRAPHIC_OBJECTS[it]->reproduct();
        if (obj)
          objects_insert(obj);
        else
          res = false;
      }
    }
    
    return res;
}

void objects_translate_start(World *w, bool no_if_glued) {
    list_item it;
    unsigned long fg, bg;
        
    forall(it, SELECTED_OBJECTS) {
      if (!no_if_glued || (no_if_glued && !GRAPHIC_OBJECTS[it]->glued())) {
        GRAPHIC_OBJECTS[it]->get_color(fg, bg);
        GRAPHIC_OBJECTS[it]->set_color(w, get_pixel_from_value(WHITE), get_pixel_from_value(WHITE));
        GRAPHIC_OBJECTS[it]->draw(w, CopyMode);
        GRAPHIC_OBJECTS[it]->set_color(w, fg, bg);
      }
    }

    objects_repaint_unselected(w);

    forall(it, SELECTED_OBJECTS) {
      if (!no_if_glued || (no_if_glued && !GRAPHIC_OBJECTS[it]->glued())) {
        GRAPHIC_OBJECTS[it]->unselect();
        GRAPHIC_OBJECTS[it]->draw(w, XorMode);
      }
    }
}

void objects_translate_change(World *w, double dx, double dy, bool no_if_glued) {
    list_item it;
    
    forall(it, SELECTED_OBJECTS) {
      if (!no_if_glued || (no_if_glued && !GRAPHIC_OBJECTS[it]->glued())) {
        GRAPHIC_OBJECTS[it]->draw(w, XorMode);
        GRAPHIC_OBJECTS[it]->translate(dx, dy);
        GRAPHIC_OBJECTS[it]->draw(w, XorMode);
      }
    }
}

void objects_translate_end(World *w, double dx, double dy, bool no_if_glued) {
    list_item it;
        
    forall(it, SELECTED_OBJECTS) {
      if (!no_if_glued || (no_if_glued && !GRAPHIC_OBJECTS[it]->glued())) {
        GRAPHIC_OBJECTS[it]->draw(w, XorMode);
        GRAPHIC_OBJECTS[it]->translate(dx, dy);
        GRAPHIC_OBJECTS[it]->select();
        GRAPHIC_OBJECTS[it]->draw(w, CopyMode);
      }
    }
}

void objects_rotate_all(World *w) {
    list_item ngit, git, sit;
    GraphicObj *obj;
    
    if (!GRAPHIC_OBJECTS.empty()) {
      GRAPHIC_OBJECTS.init_iterator();
      ngit = GRAPHIC_OBJECTS.append(obj = GRAPHIC_OBJECTS.del_item(git = GRAPHIC_OBJECTS.first()));
    
      if (obj->selected()) {	// e' necessario ajustar a lista de objetos selecionados
        SELECTED_OBJECTS.init_iterator();
        sit = SELECTED_OBJECTS.search(git);
        SELECTED_OBJECTS.assign(sit, ngit);
      }
      objects_repaint(w);  
    }
}

void objects_selected_top(World *w) {
    list_item git, it;
    bool did_something = false;
    
    forall_list_items(it, SELECTED_OBJECTS) {
      GRAPHIC_OBJECTS.init_iterator();
      git = GRAPHIC_OBJECTS.push(GRAPHIC_OBJECTS.del_item(SELECTED_OBJECTS[it]));
      SELECTED_OBJECTS.assign(it, git);
      did_something = true;
    }      
    
    if (did_something)
      objects_repaint(w);
}

void objects_selected_bottom(World *w) {
    list_item git, it;
    bool did_something = false;
    
    forall_list_items(it, SELECTED_OBJECTS) {
      GRAPHIC_OBJECTS.init_iterator();
      git = GRAPHIC_OBJECTS.append(GRAPHIC_OBJECTS.del_item(SELECTED_OBJECTS[it]));
      SELECTED_OBJECTS.assign(it, git);
      did_something = true;
    }      
    
    if (did_something)
      objects_repaint(w);
}

list(GraphicObj_ptr)& objects_get_graphic_objects() {
    return GRAPHIC_OBJECTS;
}

list(list_item)& objects_get_selected_objects() {
    return SELECTED_OBJECTS;
}

void objects_install_created_objects(World *w, bool select, Obj_ptr obj, char *function, char *module) {
    GraphicObj *graphic_obj;
    unsigned long fg, bg;
    int lw, ls;
    
    graphic_obj = obj->new_GraphicObj();
    
    // Isto 'e necessario porque o construtor da classe GraphicObj
    // n~ao ajusta as cores e atributos de linha no GC por falta do world.
    // Gon (20/08/92).
    graphic_obj->get_color(fg, bg);
    graphic_obj->set_color(w, fg, bg);
    graphic_obj->get_line_attr(lw, ls);
    graphic_obj->set_line_attr(w, lw, ls);
    
    if (select)
      graphic_obj->select();
    graphic_obj->draw(w, CopyMode);
    if (graphic_obj->obj_volatile())
      delete graphic_obj;		// consequently obj
    else {
      if (!geolab_knows_object(graphic_obj->type()))
        graphic_obj->set_external(function, module);
      if (select) {
        SELECTED_OBJECTS.init_iterator();
        SELECTED_OBJECTS.push(GRAPHIC_OBJECTS.push(graphic_obj));
      }
      else
        GRAPHIC_OBJECTS.push(graphic_obj);
    }
    
    if (select) {
      if (objects_selected_empty())
        enable_all_algorithms();
      else
        select_applyable_algorithms();

      select_applyable_mode(w);
    }
}

void objects_install_created_objects(World *w, bool select, list(Obj_ptr) &Objects, char *function, char *module) {
    Obj *obj;
    GraphicObj *graphic_obj;
    unsigned long fg, bg;
    int lw, ls;

    forall(obj, Objects) {
      graphic_obj = obj->new_GraphicObj();
      // Isto 'e necessario porque o construtor da classe GraphicObj
      // n~ao ajusta as cores e atributos de linha no GC por falta do world.
      // Gon (20/08/92).
      graphic_obj->get_color(fg, bg);
      graphic_obj->set_color(w, fg, bg);
      graphic_obj->get_line_attr(lw, ls);
      graphic_obj->set_line_attr(w, lw, ls);
      
      if (select)
        graphic_obj->select();
      graphic_obj->draw(w, CopyMode);
      if (graphic_obj->obj_volatile())
        delete graphic_obj;		// consequently obj
      else {
	  if (!geolab_knows_object(graphic_obj->type())) {
	      graphic_obj->set_external(function, module);
	  }
	  if (select) {
	      SELECTED_OBJECTS.init_iterator();
	      SELECTED_OBJECTS.push(GRAPHIC_OBJECTS.push(graphic_obj));
	  }
	  else
	    GRAPHIC_OBJECTS.push(graphic_obj);
      }
    }
    if (select) {
      if (objects_selected_empty())
        enable_all_algorithms();
      else
        select_applyable_algorithms();
        
      select_applyable_mode(w);
    }
    Objects.clear();
}

void objects_install_created_objects_no_refresh(World *w, bool select, list(Obj_ptr) &Objects, char *function, char *module) {
    Obj *obj;
    GraphicObj *graphic_obj;
    unsigned long fg, bg;
    int lw, ls;

    forall(obj, Objects) {
      graphic_obj = obj->new_GraphicObj();
      // Isto 'e necessario porque o construtor da classe GraphicObj
      // n~ao ajusta as cores e atributos de linha no GC por falta do world.
      // Gon (20/08/92).
      graphic_obj->get_color(fg, bg);
      graphic_obj->set_color(w, fg, bg);
      graphic_obj->get_line_attr(lw, ls);
      graphic_obj->set_line_attr(w, lw, ls);

      if (select)
        graphic_obj->select();
      if (graphic_obj->obj_volatile())
        delete graphic_obj;		// consequently obj
      else {
        if (!geolab_knows_object(graphic_obj->type()))
          graphic_obj->set_external(function, module);
        if (select) {
          SELECTED_OBJECTS.init_iterator();
          SELECTED_OBJECTS.push(GRAPHIC_OBJECTS.push(graphic_obj));
        }
        else
          GRAPHIC_OBJECTS.push(graphic_obj);	
      }
    }

    Objects.clear();
}


list(Obj_ptr) extract_selected_objects(ObjType type, ObjType subtype) {
    list(Obj_ptr) res;
    list_item it;
    Obj_ptr obj;
    forall(it, SELECTED_OBJECTS) {
	obj = GRAPHIC_OBJECTS[it]->obj;
	if (obj->type() == type && obj->subtype() == subtype)
	  res.push(obj);
    }
    return res;
}


list(Obj_ptr) extract_objects(ObjType type, ObjType subtype) {
    list(Obj_ptr) res;
    GraphicObj_ptr it;
    forall(it, GRAPHIC_OBJECTS) {
	if (it->obj->type() == type && it->obj->subtype() == subtype)
	  res.push(it->obj);
    }
    return res;
}


list(GraphicObj_ptr) extract_selected_graphic_objects(ObjType type, ObjType subtype) {
    list(GraphicObj_ptr) res;
    list_item it;
    Obj_ptr obj;
    forall(it, SELECTED_OBJECTS) {
	obj = GRAPHIC_OBJECTS[it]->obj;
	if (obj->type() == type && obj->subtype() == subtype)
	  res.push(GRAPHIC_OBJECTS[it]);
    }
    return res;
}


list(GraphicObj_ptr) extract_graphic_objects(ObjType type, ObjType subtype) {
    list(GraphicObj_ptr) res;
    GraphicObj_ptr it;
    forall(it, GRAPHIC_OBJECTS) {
	if (it->obj->type() == type && it->obj->subtype() == subtype)
	  res.push(it);
    }
    return res;
}




