/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 * voronoi.C - implementacao do algoritmo para calculo da triangulacao de
 *             Delaunay e do diagrama de Voronoi atraves do algoritmo
 *             apresentado por Guibas & Stolfi em "Primitives for the
 *             Manipulation of General Subdivisions and the Computation
 *             of Voronoi Diagrams", ACM TG 85
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    20.02.92  13:40 	Welson R. Jacometti  	Arquivo criado
 *
 */

#include "voronoi.h"
#include "Line2D.h"

/*
 **** includes e variaveis para animacao ***/

#include "Segment2D.h"
#include "Arrow2D.h"
#include "Circle.h"

static GraphicSegment2D *anim_s;
static GraphicArrow2D *anim_a;
static GraphicLine2D *anim_l;
static GraphicCircle *anim_c;

#define anim_segment	((Segment2D *)anim_s->obj)
#define anim_arrow	((Arrow2D *)anim_a->obj)
#define anim_line	((Line2D *)anim_l->obj)
#define anim_circle	((Circle *)anim_c->obj)

void voronoi_animation_init(World *w) {
    static bool FIRST = true;
    
    if (FIRST) {
      anim_s = new GraphicSegment2D();
      anim_a = new GraphicArrow2D();
      anim_l = new GraphicLine2D();
      anim_c = new GraphicCircle();
      FIRST = false;
    }
    anim_s->set_all(w, FOREGROUND, BACKGROUND, 0, 0, false);
    anim_a->set_all(w, FOREGROUND, BACKGROUND, 0, 0, false);
    anim_l->set_all(w, FOREGROUND, BACKGROUND, 0, 0, false);
    anim_c->set_all(w, FOREGROUND, BACKGROUND, 0, 0, false);
}

/*
 *******************************************/

static Point2D cc;		// util no calculo dos vertices do DV

void voronoi_setup_rays(QEdge_ptr e, bool FARTHEST) {
    QEdge_ptr i;
    Point2D p, int1, int2;
    Line2D dline, normal;
    SideType side;
    
    i = e;
    do {
      p = middle(*(i->rot()->org()), *(i->rot()->dest()));	// calcula o ponto medio do segmento dual (DD)
      dline.set(*(i->rot()->org()), *(i->rot()->dest()));	// calcula a equacao parametrica da reta que contem a aresta dual
      normal = dline.normal_through_point(p);			// calcula a reta normal a dline que passa por p
      
      if (!i->dest()->at_infinity())
        p = *(i->dest());
        
      if (normal.circle_intersections(p, 1000, int1, int2) != Intersect)
        error("voronoi.C", "voronoi_setup_rays", "could not get intersection points");

      side = i->rot()->org()->which_side(p, int1);
      if (side == Left || side == Colinear)
        if (FARTHEST)
          i->org()->set_at_infinity(int2.X(), int2.Y()); 
        else
          i->org()->set_at_infinity(int1.X(), int1.Y()); 
      else
        if (FARTHEST)
          i->org()->set_at_infinity(int1.X(), int1.Y());
        else
          i->org()->set_at_infinity(int2.X(), int2.Y());
 
      i = i->onext();
    } while (i != e);
}

QuadEdge_forall_bf_BEGIN(voronoi_setup_vertices,e)
    if (!already) {
      cc = circumcentre(*(e->rot()->org()), *(e->rot()->lnext()->org()), *(e->rot()->lnext()->lnext()->org()));
      already = true;
    }
    eaux->org()->setXY(cc.X(), cc.Y());
QuadEdge_forall_bf_END(voronoi_setup_vertices,e)

void setup_voronoi_geometry(QuadEdge *q, bool FARTHEST) {
    QEdge_ptr e = q->root()->rot();
    QEdge_ptr i = e;
        
    rec_flag = *(e->oflag());
    do {
      i->org()->set_at_infinity(0, 0);
      *(i->oflag()) = !rec_flag;
      i = i->onext();
    } while (i != e);
    voronoi_setup_vertices(e->sym());
    voronoi_setup_rays(e, FARTHEST);
}

TwoEdges main_algorithm(array(Point2D_ptr) &points, int first, int last, bool FARTHEST, World *w, int animation_depth, int reference_depth) {
    int npoints = last - first + 1;

    ANIMATION_BEGIN(w, animation_depth, reference_depth)
      anim_s->set_string("");
      anim_a->set_string("");
    ANIMATION_END(w, animation_depth, reference_depth)
    
    if (npoints == 2) {
      QEdge_ptr a;
      Point2D s1, s2;

      s1 = *points[first];
      s2 = *points[first + 1];

      a = new QEdge(s1, s2);

      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        anim_segment->set(s1, s2);
        anim_s->draw(w, XorMode);
      ANIMATION_END(w, animation_depth, reference_depth)
            
      return TwoEdges(a, a->sym());
    }      
    else
      if (npoints == 3) {
      	QEdge_ptr a, b, c;
      	Point2D s1, s2, s3;
      	SideType side;
      	
      	s1 = *points[first];
      	s2 = *points[first + 1];
      	s3 = *points[first + 2];

        side = s3.which_side(s1, s2);        
        
        if (FARTHEST && side == Colinear) {
          a = new QEdge(s1, s3);

          ANIMATION_BEGIN(w, animation_depth, reference_depth)
            anim_segment->set(s1, s2);
            anim_s->draw(w, XorMode);
          ANIMATION_END(w, animation_depth, reference_depth)

          return TwoEdges(a, a->sym());          
        }
        
      	a = new QEdge(s1, s2);
      	b = new QEdge(s2, s3);

      	splice(a->sym(), b);

        ANIMATION_BEGIN(w, animation_depth, reference_depth)
          anim_segment->set(s1, s2);
          anim_s->draw(w, XorMode);
          anim_segment->set(s2, s3);
          anim_s->draw(w, XorMode);
        ANIMATION_END(w, animation_depth, reference_depth)
              	
      	if (side == Left) {
      	  c = connect_edges(b, a);

          ANIMATION_BEGIN(w, animation_depth, reference_depth)
            anim_segment->set(s1, s3);
            anim_s->draw(w, XorMode);
          ANIMATION_END(w, animation_depth, reference_depth)

      	  return TwoEdges(a, b->sym());
      	}
      	else
      	  if (side == Right) {
      	    c = connect_edges (b, a);

            ANIMATION_BEGIN(w, animation_depth, reference_depth)
              anim_segment->set(s1, s3);
              anim_s->draw(w, XorMode);
            ANIMATION_END(w, animation_depth, reference_depth)

      	    return TwoEdges(c->sym(), c);
      	  }

        return TwoEdges(a, b->sym());		// s1, s2 e s3 colineares
      }	    
      else {
      	
      	/* npoints maior ou igual a 4 */
      	
      	QEdge_ptr ldo, ldi, rdo, rdi, auxl, auxr;
      	QEdge_ptr basel, lcand, rcand, t;
     	QEdge_ptr uldi, urdi;			// usados somente em FARTHEST
     	Point2D ubaselOrg, ubaselDest;
        bool firstl, firstr; 

        ANIMATION_BEGIN(w, animation_depth+1, reference_depth)
          double xmid = (points[(first + last) / 2]->X() + points[((first + last) / 2) + 1]->X()) / 2;
          anim_line->set(xmid, 0, xmid, 1);
          anim_l->draw(w, XorMode);
          frame_stop_mark(w);
        ANIMATION_END(w, animation_depth+1, reference_depth)
                
      	TwoEdges result_left  = main_algorithm(points, first, (first + last) / 2, FARTHEST, w, animation_depth, reference_depth);
      	TwoEdges result_right = main_algorithm(points, ((first + last) / 2) + 1, last, FARTHEST, w, animation_depth, reference_depth);
      	
        ANIMATION_BEGIN(w, animation_depth+1, reference_depth)
          double xmid = (points[(first + last) / 2]->X() + points[((first + last) / 2) + 1]->X()) / 2;
          anim_line->set(xmid, 0, xmid, 1);
          anim_l->draw(w, XorMode);
        ANIMATION_END(w, animation_depth+1, reference_depth)

      	ldo = result_left.le;
      	ldi = result_left.re;
      	
      	rdi = result_right.le;
      	rdo = result_right.re;

        if (FARTHEST) {				// calcula linha de suporte
          uldi = ldi->onext();			// superior (analogo ao calculo
          urdi = rdi->oprev();			// da linha inferior)
          
          ANIMATION_BEGIN(w, animation_depth, reference_depth)
            anim_segment->set(*urdi->org(), *urdi->dest());
            anim_s->draw(w, XorMode);
            anim_segment->set(*uldi->org(), *uldi->dest());
            anim_s->draw(w, XorMode);
            anim_arrow->set(*urdi->org(), *urdi->dest());
            if (animation(animation_depth+5, reference_depth))
              anim_a->set_string("urdi");
            anim_a->draw(w, XorMode);
            anim_arrow->set(*uldi->org(), *uldi->dest());
            if (animation(animation_depth+5, reference_depth))
              anim_a->set_string("uldi");
            anim_a->draw(w, XorMode);
            frame_stop_mark(w);
          ANIMATION_END(w, animation_depth, reference_depth)

          auxl = 0; firstl = true;
          auxr = 0; firstr = true;
          
          do {
            if (!left_of(urdi->org(), uldi) && auxl != uldi->rprev()) {

              ANIMATION_BEGIN(w, animation_depth, reference_depth)
                anim_arrow->set(*uldi->org(), *uldi->dest());
                if (animation(animation_depth+5, reference_depth))
                  anim_a->set_string("uldi");
                anim_a->draw(w, XorMode);
                anim_segment->set(*uldi->org(), *uldi->dest());
                anim_s->draw(w, XorMode);
              ANIMATION_END(w, animation_depth, reference_depth)

              if (firstl) {
              	firstl = false;
              	auxl = uldi;
              }
              
              uldi = uldi->rprev();

              ANIMATION_BEGIN(w, animation_depth, reference_depth)
                anim_segment->set(*uldi->org(), *uldi->dest());
                anim_s->draw(w, XorMode);
                anim_arrow->set(*uldi->org(), *uldi->dest());
                anim_a->draw(w, XorMode);
                frame_stop_mark(w);
              ANIMATION_END(w, animation_depth, reference_depth)

            }
            else
              if (!right_of(uldi->org(), urdi) && auxr != urdi->lnext()) {
 
                ANIMATION_BEGIN(w, animation_depth, reference_depth)
                  anim_arrow->set(*urdi->org(), *urdi->dest());
                  if (animation(animation_depth+5, reference_depth))
                    anim_a->set_string("urdi");
                  anim_a->draw(w, XorMode);
                  anim_segment->set(*urdi->org(), *urdi->dest());
                  anim_s->draw(w, XorMode);
                ANIMATION_END(w, animation_depth, reference_depth)

                if (firstr) {
                  firstr = false;
                  auxr = urdi;
                }
              
                urdi = urdi->lnext();

                ANIMATION_BEGIN(w, animation_depth, reference_depth)
                  anim_segment->set(*urdi->org(), *urdi->dest());
                  anim_s->draw(w, XorMode);
                  anim_arrow->set(*urdi->org(), *urdi->dest());
                  anim_a->draw(w, XorMode);
                  frame_stop_mark(w);
                ANIMATION_END(w, animation_depth, reference_depth)

              }
              else
                break;
          } while (true);

 	  ubaselOrg = *(urdi->sym()->dest());
 	  ubaselDest = *(uldi->org());
 	  
 	  ANIMATION_BEGIN(w, animation_depth, reference_depth)
            anim_arrow->set(*urdi->org(), *urdi->dest());
            if (animation(animation_depth+5, reference_depth))
              anim_a->set_string("urdi");
            anim_a->draw(w, XorMode);
            anim_arrow->set(*uldi->org(), *uldi->dest());
            if (animation(animation_depth+5, reference_depth))
              anim_a->set_string("uldi");
            anim_a->draw(w, XorMode);
            anim_segment->set(*urdi->org(), *urdi->dest());
            anim_s->draw(w, XorMode);
            anim_segment->set(*uldi->org(), *uldi->dest());
            anim_s->draw(w, XorMode);
            anim_segment->set(ubaselOrg, ubaselDest);
            if (animation(animation_depth+5, reference_depth))
              anim_s->set_string("ubasel");
            anim_s->draw(w, XorMode);
            anim_s->set_string("");
            frame_stop_mark(w);
 	  ANIMATION_END(w, animation_depth, reference_depth)
 	  
        }

        ANIMATION_BEGIN(w, animation_depth, reference_depth)
          anim_segment->set(*rdi->org(), *rdi->dest());
          anim_s->draw(w, XorMode);
          anim_segment->set(*ldi->org(), *ldi->dest());
          anim_s->draw(w, XorMode);
          anim_arrow->set(*rdi->org(), *rdi->dest());
          if (animation(animation_depth+5, reference_depth))
            anim_a->set_string("rdi");
          anim_a->draw(w, XorMode);
          anim_arrow->set(*ldi->org(), *ldi->dest());
          if (animation(animation_depth+5, reference_depth))
            anim_a->set_string("ldi");
          anim_a->draw(w, XorMode);
          frame_stop_mark(w);
        ANIMATION_END(w, animation_depth, reference_depth)
                      	
        auxl = 0; firstl = true;
        auxr = 0; firstr = true;	// calcula linha de suporte
          				// inferior				
      	do {
      	  if (FARTHEST ? (!right_of(rdi->org(), ldi) && auxl != ldi->lnext()) : left_of(rdi->org(), ldi)) {
      	  
            ANIMATION_BEGIN(w, animation_depth, reference_depth)
              anim_arrow->set(*ldi->org(), *ldi->dest());
              if (animation(animation_depth+5, reference_depth))
                anim_a->set_string("ldi");
              anim_a->draw(w, XorMode);
              anim_segment->set(*ldi->org(), *ldi->dest());
              anim_s->draw(w, XorMode);
            ANIMATION_END(w, animation_depth, reference_depth)

            if (firstl) {
              auxl = ldi;
              firstl = false;
            }             
            
      	    ldi = ldi->lnext();

            ANIMATION_BEGIN(w, animation_depth, reference_depth)
              anim_segment->set(*ldi->org(), *ldi->dest());
              anim_s->draw(w, XorMode);
              anim_arrow->set(*ldi->org(), *ldi->dest());
              anim_a->draw(w, XorMode);
              frame_stop_mark(w);
            ANIMATION_END(w, animation_depth, reference_depth)

      	  }
      	  else
      	    if (FARTHEST ? (!left_of(ldi->org(), rdi) && auxr != rdi->rprev()) : right_of(ldi->org(), rdi)) {
 
              ANIMATION_BEGIN(w, animation_depth, reference_depth)
                anim_arrow->set(*rdi->org(), *rdi->dest());
                if (animation(animation_depth+5, reference_depth))
                  anim_a->set_string("rdi");
                anim_a->draw(w, XorMode);
                anim_segment->set(*rdi->org(), *rdi->dest());
                anim_s->draw(w, XorMode);
              ANIMATION_END(w, animation_depth, reference_depth)

              if (firstr) {
                auxr = rdi;
                firstr = false;
              }             
            
      	      rdi = rdi->rprev();

              ANIMATION_BEGIN(w, animation_depth, reference_depth)
                anim_segment->set(*rdi->org(), *rdi->dest());
                anim_s->draw(w, XorMode);
                anim_arrow->set(*rdi->org(), *rdi->dest());
                anim_a->draw(w, XorMode);
                frame_stop_mark(w);
              ANIMATION_END(w, animation_depth, reference_depth)

      	    }
      	    else
      	      break;
      	} while (true);

        basel = connect_edges(rdi->sym(), ldi);

        ANIMATION_BEGIN(w, animation_depth, reference_depth)
          anim_arrow->set(*rdi->org(), *rdi->dest());
          if (animation(animation_depth+5, reference_depth))
            anim_a->set_string("rdi");
          anim_a->draw(w, XorMode);
          anim_arrow->set(*ldi->org(), *ldi->dest());
          if (animation(animation_depth+5, reference_depth))
            anim_a->set_string("ldi");
          anim_a->draw(w, XorMode);
          anim_segment->set(*rdi->org(), *rdi->dest());
          anim_s->draw(w, XorMode);
          anim_segment->set(*ldi->org(), *ldi->dest());
          anim_s->draw(w, XorMode);
          anim_segment->set(*basel->org(), *basel->dest());
          if (animation(animation_depth+5, reference_depth))
            anim_s->set_string("basel");
          anim_s->draw(w, XorMode);
          anim_s->set_string("");
          frame_stop_mark(w);
        ANIMATION_END(w, animation_depth, reference_depth)

        if (*(ldi->org()) == *(ldo->org()))
          ldo = basel->sym();
          
        if (*(rdi->org()) == *(rdo->org()))
          rdo = basel;

        if (FARTHEST) {
        	
   	  /*
  	   * Elimina os pontos que situam-se entre as duas retas de suporte e
 	   * portanto NAO sao pontos da envoltoria convexa e portanto nao fazem
 	   * parte da triangulacao final.
 	   *
 	   */

 	  lcand = basel->rprev()->sym()->onext();
 	  rcand = basel->oprev()->sym()->oprev();
 	  
 	  while (!(*(lcand->org()) == ubaselDest)) {
 	    t = lcand->sym()->onext();

            ANIMATION_BEGIN(w, animation_depth, reference_depth)
              QEdge_ptr step = lcand;
              do {
                anim_segment->set(*step->org(), *step->dest());
                anim_s->draw(w, XorMode);
                if (animation(animation_depth+5, reference_depth))
                  anim_s->set_string("lcand");
                anim_s->draw(w, XorMode);
                frame_stop_mark(w);
                anim_s->draw(w, XorMode);
                anim_s->set_string("");
                step = step->onext();
              } while (step != lcand);
            ANIMATION_END(w, animation_depth, reference_depth)

 	    delete_incident_edges(lcand);
 	    lcand = t;
 	  }
 	  
 	  while (!(*(rcand->org()) == ubaselOrg)) {
 	    t = rcand->sym()->oprev();

            ANIMATION_BEGIN(w, animation_depth, reference_depth)
              QEdge_ptr step = rcand;
              do {
                anim_segment->set(*step->org(), *step->dest());
                anim_s->draw(w, XorMode);
                if (animation(animation_depth+5, reference_depth))
                  anim_s->set_string("rcand");
                anim_s->draw(w, XorMode);
                frame_stop_mark(w);
                anim_s->draw(w, XorMode);
                anim_s->set_string("");
                step = step->onext();
              } while (step != rcand);
            ANIMATION_END(w, animation_depth, reference_depth)

 	    delete_incident_edges(rcand);
 	    rcand = t;
 	  }

 	  ANIMATION_BEGIN(w, animation_depth, reference_depth)
            anim_segment->set(ubaselOrg, ubaselDest);
            if (animation(animation_depth+5, reference_depth))
              anim_s->set_string("ubasel");
            anim_s->draw(w, XorMode);
            anim_s->set_string("");
 	  ANIMATION_END(w, animation_depth, reference_depth)
 	  
 	}

        do {
          lcand = basel->rprev();

          ANIMATION_BEGIN(w, animation_depth+2, reference_depth)
            Point2D cc = circumcentre(*(basel->org()), *(basel->dest()), *(lcand->dest()));
            double radius = cc.e_distance(*(basel->org())); 
            anim_circle->set(cc.X(), cc.Y(), radius);
            anim_c->draw(w, XorMode);
            anim_circle->set(lcand->onext()->dest()->X(), lcand->onext()->dest()->Y(), 10);
            anim_c->draw(w, XorMode);
            frame_stop_mark(w);
          ANIMATION_END(w, animation_depth+2, reference_depth)

          if (valid(lcand, basel))
            while (lcand->onext()->dest()->inCircle(*(basel->dest()), *(basel->org()), *(lcand->dest())) == (FARTHEST ? Outside : Inside)) {
              t = lcand->onext();

              ANIMATION_BEGIN(w, animation_depth, reference_depth)
                anim_segment->set(*lcand->org(), *lcand->dest());
                anim_s->draw(w, XorMode);
                if (animation(animation_depth+5, reference_depth))
                  anim_s->set_string("lcand");
                anim_s->draw(w, XorMode);
                frame_stop_mark(w);
                anim_s->draw(w, XorMode);
                anim_s->set_string("");
              ANIMATION_END(w, animation_depth, reference_depth)

              ANIMATION_BEGIN(w, animation_depth+2, reference_depth)
                Point2D cc = circumcentre(*(basel->org()), *(basel->dest()), *(lcand->dest()));
                double radius = cc.e_distance(*(basel->org())); 
                anim_circle->set(cc.X(), cc.Y(), radius);
                anim_c->draw(w, XorMode);
                anim_circle->set(lcand->onext()->dest()->X(), lcand->onext()->dest()->Y(), 10);
                anim_c->draw(w, XorMode);
              ANIMATION_END(w, animation_depth+2, reference_depth)

              delete lcand; 
              lcand = t;

              ANIMATION_BEGIN(w, animation_depth+2, reference_depth)
                Point2D cc = circumcentre(*(basel->org()), *(basel->dest()), *(lcand->dest()));
                double radius = cc.e_distance(*(basel->org())); 
                anim_circle->set(cc.X(), cc.Y(), radius);
                anim_c->draw(w, XorMode);
                anim_circle->set(lcand->onext()->dest()->X(), lcand->onext()->dest()->Y(), 10);
                anim_c->draw(w, XorMode);
                frame_stop_mark(w);
              ANIMATION_END(w, animation_depth+2, reference_depth)
    	    
            }

          ANIMATION_BEGIN(w, animation_depth+2, reference_depth)
            Point2D cc = circumcentre(*(basel->org()), *(basel->dest()), *(lcand->dest()));
            double radius = cc.e_distance(*(basel->org())); 
            anim_circle->set(cc.X(), cc.Y(), radius);
            anim_c->draw(w, XorMode);
            anim_circle->set(lcand->onext()->dest()->X(), lcand->onext()->dest()->Y(), 10);
            anim_c->draw(w, XorMode);
          ANIMATION_END(w, animation_depth+2, reference_depth)

          rcand = basel->oprev();
          
          ANIMATION_BEGIN(w, animation_depth+2, reference_depth)
            Point2D cc = circumcentre(*(basel->org()), *(basel->dest()), *(rcand->dest()));
            double radius = cc.e_distance(*(basel->org())); 
            anim_circle->set(cc.X(), cc.Y(), radius);
            anim_c->draw(w, XorMode);
            anim_circle->set(rcand->oprev()->dest()->X(), rcand->oprev()->dest()->Y(), 10);
            anim_c->draw(w, XorMode);
            frame_stop_mark(w);
          ANIMATION_END(w, animation_depth+2, reference_depth)
    	    
          if (valid(rcand, basel))
            while (rcand->oprev()->dest()->inCircle(*(basel->dest()), *(basel->org()), *(rcand->dest())) == (FARTHEST ? Outside : Inside)) {
              t = rcand->oprev();

              ANIMATION_BEGIN(w, animation_depth, reference_depth)
                anim_segment->set(*rcand->org(), *rcand->dest());
                anim_s->draw(w, XorMode);
                if (animation(animation_depth+5, reference_depth))
                  anim_s->set_string("rcand");
                anim_s->draw(w, XorMode);
                frame_stop_mark(w);
                anim_s->draw(w, XorMode);
                anim_s->set_string("");
              ANIMATION_END(w, animation_depth, reference_depth)

              ANIMATION_BEGIN(w, animation_depth+2, reference_depth)
                Point2D cc = circumcentre(*(basel->org()), *(basel->dest()), *(rcand->dest()));
                double radius = cc.e_distance(*(basel->org())); 
                anim_circle->set(cc.X(), cc.Y(), radius);
                anim_c->draw(w, XorMode);
                anim_circle->set(rcand->oprev()->dest()->X(), rcand->oprev()->dest()->Y(), 10);
                anim_c->draw(w, XorMode);
              ANIMATION_END(w, animation_depth+2, reference_depth)
    	    
              delete rcand; 
              rcand = t;

              ANIMATION_BEGIN(w, animation_depth+2, reference_depth)
                Point2D cc = circumcentre(*(basel->org()), *(basel->dest()), *(rcand->dest()));
                double radius = cc.e_distance(*(basel->org())); 
                anim_circle->set(cc.X(), cc.Y(), radius);
                anim_c->draw(w, XorMode);
                anim_circle->set(rcand->oprev()->dest()->X(), rcand->oprev()->dest()->Y(), 10);
                anim_c->draw(w, XorMode);
                frame_stop_mark(w);
              ANIMATION_END(w, animation_depth+2, reference_depth)
    	    
            }
              	
          ANIMATION_BEGIN(w, animation_depth+2, reference_depth)
            Point2D cc = circumcentre(*(basel->org()), *(basel->dest()), *(rcand->dest()));
            double radius = cc.e_distance(*(basel->org())); 
            anim_circle->set(cc.X(), cc.Y(), radius);
            anim_c->draw(w, XorMode);
            anim_circle->set(rcand->oprev()->dest()->X(), rcand->oprev()->dest()->Y(), 10);
            anim_c->draw(w, XorMode);
          ANIMATION_END(w, animation_depth+2, reference_depth)
    	    
    	  if (!valid(lcand, basel) && !valid(rcand, basel))
    	    break;
    	    
    	  if (!valid(lcand, basel) || (valid(rcand, basel) &&  rcand->dest()->inCircle(*(lcand->dest()), *(lcand->org()), *(rcand->org())) == (FARTHEST ? Outside : Inside))) {
    	   
    	    ANIMATION_BEGIN(w, animation_depth+1, reference_depth)
    	      Point2D cc = circumcentre(*(basel->org()), *(basel->dest()), *(rcand->dest()));
              double radius = cc.e_distance(*(basel->org())); 
    	      anim_circle->set(cc.X(), cc.Y(), radius);
    	      anim_c->draw(w, XorMode);
    	    ANIMATION_END(w, animation_depth+1, reference_depth)
    	    
            ANIMATION_BEGIN(w, animation_depth+5, reference_depth)    	    
              anim_segment->set(*basel->org(), *basel->dest());
              anim_s->set_string("basel");
              anim_s->draw(w, XorMode);
              anim_s->set_string("");
              anim_s->draw(w, XorMode);
            ANIMATION_END(w, animation_depth+5, reference_depth)    	    
            
    	    basel = connect_edges(rcand, basel->sym());
    	   
    	    ANIMATION_BEGIN(w, animation_depth+1, reference_depth)
    	      frame_stop_mark(w);
    	      anim_c->draw(w, XorMode);
    	    ANIMATION_END(w, animation_depth+1, reference_depth)
    	    
    	  }
    	  else {
    	   
    	    ANIMATION_BEGIN(w, animation_depth+1, reference_depth)
    	      Point2D cc = circumcentre(*(basel->org()), *(basel->dest()), *(lcand->dest()));
              double radius = cc.e_distance(*(basel->org())); 
    	      anim_circle->set(cc.X(), cc.Y(), radius);
    	      anim_c->draw(w, XorMode);
    	    ANIMATION_END(w, animation_depth+1, reference_depth)

            ANIMATION_BEGIN(w, animation_depth+5, reference_depth)    	    
              anim_segment->set(*basel->org(), *basel->dest());
              anim_s->set_string("basel");
              anim_s->draw(w, XorMode);
              anim_s->set_string("");
              anim_s->draw(w, XorMode);
            ANIMATION_END(w, animation_depth+5, reference_depth)    	    
            
    	    basel = connect_edges(basel->sym(), lcand->sym());
    	   
    	    ANIMATION_BEGIN(w, animation_depth+1, reference_depth)
    	      frame_stop_mark(w);
    	      anim_c->draw(w, XorMode);
    	    ANIMATION_END(w, animation_depth+1, reference_depth)
    	    
    	  }

          ANIMATION_BEGIN(w, animation_depth, reference_depth)
            anim_segment->set(*basel->org(), *basel->dest());
            if (animation(animation_depth+5, reference_depth))
              anim_s->set_string("basel");
            anim_s->draw(w, XorMode);
            anim_s->set_string("");
            frame_stop_mark(w);
          ANIMATION_END(w, animation_depth, reference_depth)

        } while (true);

        ANIMATION_BEGIN(w, animation_depth+5, reference_depth)    	    
          anim_segment->set(*basel->org(), *basel->dest());
          anim_s->set_string("basel");
          anim_s->draw(w, XorMode);
          anim_s->set_string("");
          anim_s->draw(w, XorMode);
        ANIMATION_END(w, animation_depth+5, reference_depth)    	    
            
        return TwoEdges(ldo, rdo);
      }
}    

QuadEdge* calc_triangulation(list(Point2D_ptr) &l, bool FARTHEST, World *w, int animation_depth, int reference_depth) {
    l.sort((int(*)(Point2D_ptr&, Point2D_ptr&))lexico_Xcomp_ptr);
    remove_duplicated_points(l);
    if (l.size() < 2)
      return 0;
    else {
      int i = 0;
      array(Point2D_ptr) a(0, l.size() - 1);
      Point2D_ptr p;
      forall(p, l) {
        a[i++] = p;
      }   
      
      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        voronoi_animation_init(w);
      ANIMATION_END(w, animation_depth, reference_depth)
         	
      TwoEdges aux = main_algorithm(a, 0, a.size() - 1, FARTHEST, w, animation_depth, reference_depth);
      return new QuadEdge(aux.le, aux.le, aux.re);
    }
}

QuadEdge* calc_diagram(list(Point2D_ptr) &l, bool FARTHEST, World *w, int animation_depth, int reference_depth) {
    QuadEdge *q = calc_triangulation(l, FARTHEST, w, animation_depth, reference_depth);
    if (q) {
      setup_voronoi_geometry(q, FARTHEST);
      q->set(q->root()->rot(), q->le(), q->re());
    }
    return q;
}

QuadEdge *delaunay_triangulation(list(Point2D_ptr) &points, World *w, int animation_depth, int reference_depth) {
    return calc_triangulation(points, false, w, animation_depth, reference_depth);
}

QuadEdge *voronoi_diagram(list(Point2D_ptr) &points, World *w, int animation_depth, int reference_depth) {
    return calc_diagram(points, false, w, animation_depth, reference_depth);
}

QuadEdge *farthest_neighbor_triangulation(list(Point2D_ptr) &points, World *w, int animation_depth, int reference_depth) {
    return calc_triangulation(points, true, w, animation_depth, reference_depth);
}

QuadEdge *farthest_neighbor_voronoi_diagram(list(Point2D_ptr) &points, World *w, int animation_depth, int reference_depth) {
    return calc_diagram(points, true, w, animation_depth, reference_depth);
}

// Modelagem

static list(Point2D_ptr) extract_points(const SelectedObjects &s, const GraphicObjects &obj) {
    list(Point2D_ptr) res;
    list_item it;
    forall(it, s) {
      if (obj[it]->obj->type() == Point2D_t)
        res.push((Point2D *)obj[it]->obj);
    }
    return res;
}

bool GenericVoronoiTools::is_able_to_handle(const SelectedObjects &SO, const GraphicObjects &GO) {
    list_item it;
    
    forall(it, SO) {
      if (GO[it]->type() != Point2D_t) {
      	SO.init_iterator();
      	return false;
      }
    }
    
    return true;
}

ResultObjects VoronoiDiagram::binder(const SelectedObjects &s, const GraphicObjects &obj, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    QuadEdge *voronoi;
    
    if (!s.empty()) {
      if (voronoi = calc_diagram(extract_points(s, obj), false, w, animation_depth, reference_depth)) {
        voronoi->subtype(VoronoiDiagram_t);
        result.push(voronoi);
      }
    }
    return result;
}

ResultObjects DelaunayTriangulation::binder(const SelectedObjects &s, const GraphicObjects &obj, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    QuadEdge *triang;
    
    if (!s.empty()) {
      if (triang = calc_triangulation(extract_points(s, obj), false, w, animation_depth, reference_depth)) {
        triang->subtype(DelaunayTriangulation_t);
        result.push(triang);
      }
    }

    return result;
}

ResultObjects FNVoronoiDiagram::binder(const SelectedObjects &s, const GraphicObjects &obj, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    QuadEdge *fvor;

    if (!s.empty()) {
      if (fvor = calc_diagram(extract_points(s, obj), true, w, animation_depth, reference_depth)) {
        fvor->subtype(FarthestNeighborsVoronoiDiagram_t);
        result.push(fvor);
      }
    }
    return result;
}

ResultObjects FNTriangulation::binder(const SelectedObjects &s, const GraphicObjects &obj, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    QuadEdge *ftriang;
    
    if (!s.empty()) {
      if (ftriang = calc_triangulation(extract_points(s, obj), true, w, animation_depth, reference_depth)) {
        ftriang->subtype(FarthestNeighborsTriangulation_t);
        result.push(ftriang);
      }
    }

    return result;
}

Algorithm* VoronoiDiagram_link() {
    return new VoronoiDiagram();
}

Algorithm* DelaunayTriangulation_link() {
    return new DelaunayTriangulation();
}

Algorithm* FNVoronoiDiagram_link() {
    return new FNVoronoiDiagram();
}

Algorithm* FNTriangulation_link() {
    return new FNTriangulation();
}


